import collections
import dataclasses
import re
from collections.abc import Callable
from typing import Literal, Protocol, runtime_checkable

import torch

from fme.core.typing_ import TensorDict, TensorMapping


def replace_on_mask(
    original: torch.Tensor,
    replacement: torch.Tensor,
    mask: torch.Tensor,
    mask_value: int,
):
    """Replace original with replacement in masked regions.

    Args:
        original: The original data tensor.
        replacement: The replacement data tensor.
        mask: The mask tensor.
        mask_value: The value of the mask variable in the region to be replaced.
    """
    rounded_mask = torch.round(mask).to(int)
    return torch.where(
        condition=rounded_mask == mask_value,
        input=replacement,
        other=original,
    )


@runtime_checkable
class HasGetMaskTensorFor(Protocol):
    def build_output_masker(self) -> Callable[[TensorMapping], TensorDict]: ...

    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None:
        """Get the mask for a specific variable name."""
        ...

    def to(self, device: str) -> "HasGetMaskTensorFor": ...


@dataclasses.dataclass
class StaticMaskingConfig:
    """
    Replace static masked regions with a fill value.

    Parameters:
        mask_value: Value of the mask variable in masked regions. Either 0 or 1.
        fill_value: A float fill value to use outside of masked regions. Can also be
            "mean", in which case the normalizer means are used as channel-specific
            fill values.
        exclude_names_and_prefixes: Names (2D variables) and prefixes (3D variables)
            to exclude when applying the mask.

    """

    mask_value: int
    fill_value: Literal["mean"] | float = 0.0
    exclude_names_and_prefixes: list[str] | None = None

    def __post_init__(self):
        if self.mask_value not in [0, 1]:
            raise ValueError(
                f"mask_value must be either 0 or 1, but got {self.mask_value}"
            )

    def build(self, mask: HasGetMaskTensorFor, means: TensorMapping | None = None):
        """
        Build StaticMasking.

        """
        if isinstance(self.fill_value, float):
            return StaticMasking(
                mask_value=self.mask_value,
                fill_value=collections.defaultdict(
                    lambda: torch.as_tensor(self.fill_value)
                ),
                mask=mask,
                exclude_names_and_prefixes=self.exclude_names_and_prefixes,
            )
        if means is None:
            raise ValueError(
                "fill_values mapping required by build unless configured "
                "fill_value is a float."
            )
        return StaticMasking(
            mask_value=self.mask_value,
            fill_value=means,
            mask=mask,
            exclude_names_and_prefixes=self.exclude_names_and_prefixes,
        )


class StaticMasking:
    def __init__(
        self,
        mask_value: int,
        fill_value: float | TensorMapping,
        mask: HasGetMaskTensorFor,
        exclude_names_and_prefixes: list[str] | None = None,
    ):
        if isinstance(fill_value, float):
            fill_mapping: TensorMapping = collections.defaultdict(
                lambda: torch.as_tensor(fill_value)
            )
        else:
            fill_mapping = fill_value
        self._fill_mapping = fill_mapping
        self._mask_value = mask_value
        self._mask = mask
        self._exclude_regex = self._build_regex(exclude_names_and_prefixes)

    def _build_regex(self, names_and_prefixes: list[str] | None) -> str | None:
        if names_and_prefixes:
            regex = []
            for name in names_and_prefixes:
                if name.endswith("_"):
                    regex.append(rf"^{name}\d+$")
                elif not re.match(r".+_\d+$", name):
                    regex.append(f"^{name}$")
                    regex.append(rf"^{name}_\d+$")
                else:
                    regex.append(rf"^{name}$")
            return r"|".join(regex)
        return None

    def _masks(self, name: str):
        exclude = self._exclude_regex and re.match(self._exclude_regex, name)
        return not exclude

    def __call__(self, data: TensorMapping) -> TensorDict:
        """
        Apply masking to the data for standard names recognized by a stacker.

        Args:
            data: The data to mask.

        """
        data_: TensorDict = {**data}
        for name, tensor in data_.items():
            if not self._masks(name):
                continue
            mask = self._mask.get_mask_tensor_for(name)
            if mask is None:
                continue
            try:
                fill_value = self._fill_mapping[name]
            except KeyError as err:
                raise KeyError(
                    "StaticMasking was initialized with a fill_value mapping "
                    f"but the mapping is missing key '{name}'."
                ) from err
            fill = torch.full_like(tensor, fill_value)
            mask = mask.expand(fill.shape)
            masked = replace_on_mask(
                original=tensor,
                replacement=fill,
                mask=mask,
                mask_value=self._mask_value,
            )
            data_[name] = masked
        return data_


class NullMasking:
    def __call__(self, data: TensorMapping) -> TensorDict:
        return dict(data)



import re
from collections.abc import Callable

from torch import nn


def wildcard_match(pattern: str, name: str) -> bool:
    """
    Check if a name matches a wildcard pattern.

    A wildcard pattern can include "*" to match any number of characters.
    """
    # use regex
    pattern = pattern.replace(".", r"\.")
    pattern = pattern.replace("*", ".*")
    pattern = f"^{pattern}$"
    return bool(re.match(pattern, name))


def apply_by_wildcard(
    model: nn.Module,
    func: Callable[[nn.Module, str], None],
    include: list[str],
    exclude: list[str],
):
    missing_parameters = []
    for name in model.state_dict().keys():
        if any(wildcard_match(pattern, name) for pattern in include):
            if any(wildcard_match(pattern, name) for pattern in exclude):
                raise ValueError(
                    f"Parameter {name} is included in both include "
                    f"{include} and exclude {exclude}"
                )
            func(model, name)
        elif not any(wildcard_match(pattern, name) for pattern in exclude):
            missing_parameters.append(name)
    if len(missing_parameters) > 0:
        raise ValueError(
            f"Model has parameters {missing_parameters} which are not "
            f"specified in either include {include} "
            f"or exclude {exclude}"
        )
    return model



import dataclasses
import datetime
from collections.abc import Callable
from unittest.mock import MagicMock

import torch
from torch import nn

from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.ocean import OceanConfig

from .step import StepABC, StepConfigABC, StepSelector


class MockStep(StepABC):
    def __init__(
        self,
        config: "MockStepConfig",
        dataset_info: DatasetInfo,
        init_weights: Callable[[list[nn.Module]], None],
    ):
        self.dataset_info = dataset_info
        self._config = config
        self._init_weights = init_weights

    @property
    def config(self) -> "MockStepConfig":
        return self._config

    @property
    def modules(self):
        raise NotImplementedError()

    @property
    def normalizer(self):
        raise NotImplementedError()

    @property
    def surface_temperature_name(self):
        return None

    @property
    def ocean_fraction_name(self):
        return None

    def get_regularizer_loss(self) -> torch.Tensor:
        return torch.tensor(0.0)

    def step(self, input, next_step_input_data, use_activation_checkpointing=False):
        raise NotImplementedError()

    def get_state(self):
        return {}

    def load_state(self, state):
        pass


@StepSelector.register("mock")
@dataclasses.dataclass
class MockStepConfig(StepConfigABC):
    in_names: list[str] = dataclasses.field(default_factory=list)
    out_names: list[str] = dataclasses.field(default_factory=list)

    def get_step(
        self, dataset_info: DatasetInfo, init_weights: Callable[[list[nn.Module]], None]
    ):
        return MockStep(self, dataset_info, init_weights)

    @property
    def diagnostic_names(self) -> list[str]:
        return list(set(self.out_names).difference(self.in_names))

    def get_next_step_forcing_names(self) -> list[str]:
        return []

    @property
    def input_names(self) -> list[str]:
        return self.in_names

    @property
    def output_names(self) -> list[str]:
        return self.out_names

    @property
    def next_step_input_names(self):
        raise NotImplementedError()

    @property
    def loss_names(self) -> list[str]:
        return self.out_names

    @property
    def n_ic_timesteps(self) -> int:
        raise NotImplementedError()

    def replace_ocean(self, ocean: OceanConfig | None):
        raise NotImplementedError()

    def get_ocean(self) -> OceanConfig | None:
        return None

    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ):
        raise NotImplementedError()

    def load(self):
        pass


def test_register():
    """Make sure that the registry is working as expected."""
    selector = StepSelector(type="mock", config={})
    img_shape = (16, 32)
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(7), bk=torch.arange(7)
    )
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(img_shape[0]), lon=torch.zeros(img_shape[1])
    )
    timestep = datetime.timedelta(hours=6)
    dataset_info = DatasetInfo(
        horizontal_coordinates=horizontal_coordinate,
        vertical_coordinate=vertical_coordinate,
        timestep=timestep,
    )
    init_weights = MagicMock()
    step = selector.get_step(dataset_info, init_weights)
    assert isinstance(step, MockStep)
    assert step.dataset_info == dataset_info
    assert step._init_weights == init_weights



import numpy as np
import pytest

from .plotting import (
    _stitch_data_panels,
    fold_healpix_data,
    get_cmap_limits,
    plot_imshow,
    plot_paneled_data,
)


def test_cmap_limits():
    data = np.array([1, 2, 3])
    vmin, vmax = get_cmap_limits(data)
    assert vmin == 1
    assert vmax == 3


def test_cmap_limits_diverging():
    data = np.array([-1, 2, 3])
    vmin, vmax = get_cmap_limits(data, diverging=True)
    assert vmin == -3
    assert vmax == 3


@pytest.mark.parametrize("use_colorbar", [True, False])
def test_plot_imshow(use_colorbar):
    shape = [10, 15]
    data = np.random.randn(*shape)
    fig = plot_imshow(np.array(data), use_colorbar=use_colorbar)
    width, height = (fig.get_size_inches() * fig.dpi).astype(int)
    if use_colorbar:
        # colorbar is no more than 15% of the width but greater than 0 pixels
        assert shape[1] < width <= int(shape[1] * 1.15)
        assert height == shape[0]
    else:
        assert [height, width] == shape


def test_fold_healpix_data():
    face_shape = [2, 3]
    data = np.random.randn(12, *face_shape)
    folded = fold_healpix_data(data, fill_value=0)
    expected_shape = (6 * face_shape[0], 4 * face_shape[1])
    assert folded.shape == expected_shape


@pytest.mark.parametrize("use_colorbar", [True, False])
def test_plot_imshow_healpix(use_colorbar):
    face_shape = [4, 6]
    shape = [6 * face_shape[0], 4 * face_shape[1]]
    data = np.random.randn(12, *face_shape)
    fig = plot_imshow(np.array(data), use_colorbar=use_colorbar)
    width, height = (fig.get_size_inches() * fig.dpi).astype(int)
    if use_colorbar:
        # colorbar is no more than 15% of the width but greater than 0 pixels
        assert shape[1] < width <= int(shape[1] * 1.15)
        assert height == shape[0]
    else:
        assert [height, width] == shape


def test_stitch_data_panels():
    data = [
        [np.array([[1, 2]]), np.array([[3, 4]])],
        [np.array([[5, 6]]), np.array([[7, 8]])],
    ]
    stitched = _stitch_data_panels(data, fill_value=1)
    expected = np.array(
        [  # vertical orientation is swapped as data starts from bottom-left
            [5, 6, 1, 7, 8],
            [1, 1, 1, 1, 1],
            [1, 2, 1, 3, 4],
        ]
    )
    assert np.array_equal(stitched, expected)


@pytest.mark.parametrize(
    "shape, img_shape",
    [
        pytest.param(
            [12, 2, 3],
            [
                27,  # 3 * 4 + 1 (divider) + 3 * 4 + 2 (colorbar)
                25,  # 2 * 6 + 1 (divider) + 2 * 6 + 2 (colorbar)
            ],
            id="healpix",
        ),
        pytest.param(
            [2, 3],
            [
                9,  # 3 + 1 (divider) + 3 + 2 (colorbar)
                5,  # 2 + 1 (divider) + 2 (colorbar)
            ],
            id="latlon",
        ),
    ],
)
def test_plot_paneled_data(shape, img_shape):
    panel = np.random.uniform(size=shape)
    data = [
        [panel, panel],
        [panel, panel],
    ]
    fig = plot_paneled_data(data, diverging=False)
    assert fig.image is not None
    assert np.array_equal(fig.image.size, img_shape)
    fig = plot_paneled_data(data, diverging=True)
    assert fig.image is not None
    assert np.array_equal(fig.image.size, img_shape)



import copy
import datetime

import numpy as np
import pytest
import torch
import xarray as xr

from fme.core.coordinates import (
    DimSize,
    HEALPixCoordinates,
    HorizontalCoordinates,
    LatLonCoordinates,
)

from .utils import (
    _broadcast_array_to_tensor,
    _get_indexers,
    as_broadcasted_tensor,
    decode_timestep,
    encode_timestep,
    get_horizontal_coordinates,
)

LON_DIM = "lon"
LAT_DIM = "lat"
TIME_DIM = "time"
FULL_NAME = "full"
CONSTANT_NAME = "constant"
SLICE_NONE = slice(None)


def get_sizes(
    spatial_dims: HorizontalCoordinates = LatLonCoordinates(
        lon=torch.Tensor(np.arange(6)),
        lat=torch.Tensor(np.arange(12)),
    ),
):
    spatial_sizes: list[DimSize] = copy.deepcopy(spatial_dims.loaded_sizes)
    return [DimSize(TIME_DIM, 3)] + spatial_sizes


def create_reference_dataset(
    spatial_dims: HorizontalCoordinates = LatLonCoordinates(
        lon=torch.Tensor(np.arange(6)),
        lat=torch.Tensor(np.arange(12)),
    ),
):
    dims = [TIME_DIM] + spatial_dims.dims
    dim_sizes = get_sizes(spatial_dims)
    shape = tuple(dim_size.size for dim_size in dim_sizes)
    data = np.arange(np.prod(shape)).reshape(shape)
    coords = [np.arange(size) for size in shape]
    full = xr.DataArray(data, dims=dims, coords=coords, name=FULL_NAME)
    constant = full.isel({TIME_DIM: 0}).drop_vars(TIME_DIM).rename(CONSTANT_NAME)
    return xr.merge([full, constant])


@pytest.mark.parametrize(
    "spatial_dimensions, expected_coords",
    [
        (
            "healpix",
            HEALPixCoordinates(
                face=torch.Tensor(np.arange(12)),
                width=torch.Tensor(np.arange(64)),
                height=torch.Tensor(np.arange(64)),
            ),
        ),
        (
            "latlon",
            LatLonCoordinates(
                lon=torch.Tensor(np.arange(6)),
                lat=torch.Tensor(np.arange(12)),
            ),
        ),
    ],
)
def test_get_horizontal_coordinates(spatial_dimensions, expected_coords):
    ds = create_reference_dataset(spatial_dims=expected_coords)
    result, loaded_dim_names = get_horizontal_coordinates(ds, spatial_dimensions, None)
    assert expected_coords.dims == result.dims
    for name in expected_coords.dims:
        np.testing.assert_array_equal(expected_coords.coords[name], result.coords[name])


@pytest.mark.parametrize(
    ["coordinate_type", "coord_sizes"],
    [
        pytest.param(LatLonCoordinates, {"lat": 90, "lon": 180}),
        pytest.param(HEALPixCoordinates, {"face": 12, "height": 64, "width": 64}),
    ],
)
def test_horizonal_dimension_sizes(coordinate_type, coord_sizes):
    coords = {name: torch.Tensor(np.arange(size)) for name, size in coord_sizes.items()}
    horizontal_coords = coordinate_type(**coords)
    for name, size in coord_sizes.items():
        assert len(horizontal_coords.coords[name]) == size


@pytest.mark.parametrize(
    ("variable_dims", "expected"),
    [
        ((TIME_DIM, LAT_DIM, LON_DIM), (SLICE_NONE, SLICE_NONE, SLICE_NONE)),
        ((LAT_DIM, LON_DIM), (None, SLICE_NONE, SLICE_NONE)),
        ((TIME_DIM,), (SLICE_NONE, None, None)),
        ((LAT_DIM,), (None, SLICE_NONE, None)),
        ((LON_DIM,), (None, None, SLICE_NONE)),
    ],
    ids=lambda x: f"{x}",
)
def test__get_indexers(variable_dims, expected):
    dim_sizes = get_sizes()
    shape = tuple(
        dim_size.size for dim_size in dim_sizes if dim_size.name in variable_dims
    )
    variable = xr.Variable(variable_dims, np.zeros(shape))
    dims = (TIME_DIM, LAT_DIM, LON_DIM)
    result = _get_indexers(variable, dims)
    assert result == expected


@pytest.mark.parametrize(
    "variable_dims",
    [
        (TIME_DIM, LAT_DIM, LON_DIM),
        (LAT_DIM, LON_DIM),
        (TIME_DIM,),
        (LAT_DIM,),
        (LON_DIM,),
    ],
)
def test_as_broadcasted_tensor(variable_dims):
    ds = create_reference_dataset()

    variable_shape = tuple(ds.sizes[dim] for dim in variable_dims)
    variable = xr.Variable(variable_dims, np.zeros(variable_shape))
    da = xr.DataArray(variable)

    dims = ds[FULL_NAME].dims
    shape = (ds.sizes[TIME_DIM], ds.sizes[LAT_DIM], ds.sizes[LON_DIM])
    result = as_broadcasted_tensor(variable, dims, shape)

    xarray_broadcast = da.broadcast_like(ds[FULL_NAME]).transpose(*dims)

    # Suppress read-only warning when casting to a tensor; note this is only
    # an issue following an xarray broadcast, which is used for convenience
    # in this test. The internals of the data loader convert the array to a
    # tensor prior to the broadcast, which avoids this issue.
    xarray_broadcast = xarray_broadcast.copy()

    expected = torch.as_tensor(xarray_broadcast.values)

    assert torch.equal(result, expected)


def test_encode_decode_timestep_roundtrip():
    timestep = datetime.timedelta(days=1, microseconds=1)
    encoded = encode_timestep(timestep)
    roundtripped = decode_timestep(encoded)
    assert roundtripped == timestep


def test__broadcast_array_to_tensor_no_broadcast():
    arr = np.arange(6).reshape(1, 2, 3)
    out = _broadcast_array_to_tensor(arr, (TIME_DIM, LAT_DIM, LON_DIM), (1, 2, 3))
    expected = torch.as_tensor(arr)
    torch.testing.assert_close(out, expected)


def test__broadcast_array_to_tensor_with_broadcast():
    arr = np.arange(6)
    out = _broadcast_array_to_tensor(arr, (TIME_DIM, LAT_DIM, LON_DIM), (6, 2, 3))
    expected = torch.broadcast_to(torch.as_tensor(arr)[:, None, None], (6, 2, 3))
    torch.testing.assert_close(out, expected)


def test__broadcast_array_to_tensor_raises_assertion_error():
    arr = np.zeros((1, 2))
    with pytest.raises(ValueError, match="must be 1D"):
        _broadcast_array_to_tensor(arr, (TIME_DIM, LAT_DIM, LON_DIM), (1, 2, 3))

    arr = np.zeros(3)
    with pytest.raises(ValueError, match="matching time dimension"):
        _broadcast_array_to_tensor(arr, (TIME_DIM, LAT_DIM, LON_DIM), (4, 2, 3))



import logging
from collections.abc import Mapping
from typing import Any, cast

import numpy as np
import torch
import xarray as xr

from fme.ace.aggregator.plotting import plot_paneled_data
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorMapping
from fme.core.wandb import Image


class SeasonalAggregator:
    def __init__(
        self,
        ops: GriddedOperations,
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
    ):
        self._area_weighted_mean = ops.area_weighted_mean
        self._variable_metadata = variable_metadata
        self._target_dataset: xr.Dataset | None = None
        self._gen_dataset: xr.Dataset | None = None

    @torch.no_grad()
    def record_batch(
        self,
        time: xr.DataArray,
        target_data: TensorMapping,
        gen_data: TensorMapping,
    ):
        """Record a batch of data for computing time variability statistics."""
        target_data = {name: value.cpu() for name, value in target_data.items()}
        gen_data = {name: value.cpu() for name, value in gen_data.items()}
        target_ds = _to_dataset(target_data, time)
        gen_ds = _to_dataset(gen_data, time)

        # must keep a separate dataset for each sample to avoid averaging across
        # samples when we groupby year
        if self._target_dataset is None:
            self._target_dataset = target_ds.groupby(
                target_ds.valid_time.dt.season
            ).sum(dim="stacked_sample_time", skipna=False)
        else:
            self._target_dataset = _add_dataarray(
                self._target_dataset,
                target_ds.groupby(target_ds.valid_time.dt.season).sum(
                    dim="stacked_sample_time", skipna=False
                ),
            )

        if self._gen_dataset is None:
            self._gen_dataset = gen_ds.groupby(gen_ds.valid_time.dt.season).sum(
                dim="stacked_sample_time", skipna=False
            )
        else:
            self._gen_dataset = _add_dataarray(
                self._gen_dataset,
                gen_ds.groupby(gen_ds.valid_time.dt.season).sum(
                    dim="stacked_sample_time", skipna=False
                ),
            )

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, Any]:
        if self._target_dataset is None or self._gen_dataset is None:
            raise ValueError("No data has been recorded yet.")
        dist = Distributed.get_instance()
        if dist.world_size > 1:
            target = _reduce_datasets(dist, self._target_dataset)
            gen = _reduce_datasets(dist, self._gen_dataset)
        else:
            target = self._target_dataset
            gen = self._gen_dataset
        if not dist.is_root():
            return {}
        if target is None or gen is None:
            raise RuntimeError("we are on root but no data was collected")

        if len(gen.season) < 4 or len(target.season) < 4:
            return {}  # seasonal metrics undefined when not all seasons are recorded

        target = cast(xr.Dataset, target / target["counts"])  # type: ignore
        gen = cast(xr.Dataset, gen / gen["counts"])  # type: ignore
        bias = gen - target
        plots: dict[str, Image] = {}
        metric_logs: dict[str, float] = {}

        for name in gen.data_vars.keys():
            if name == "counts":
                continue

            if self._variable_metadata is not None and name in self._variable_metadata:
                long_name = self._variable_metadata[name].long_name
                units = self._variable_metadata[name].units
                caption_name = f"{long_name} ({units})"
            else:
                caption_name = name

            target_mean_pattern = target[name].mean(dim="season")
            gen_anomaly = gen[name] - target_mean_pattern
            target_anomaly = target[name] - target_mean_pattern
            r2 = get_r2(gen_anomaly, target_anomaly)

            image = plot_paneled_data(
                [
                    [
                        target_anomaly.sel(season="DJF").values,
                        target_anomaly.sel(season="MAM").values,
                        target_anomaly.sel(season="JJA").values,
                        target_anomaly.sel(season="SON").values,
                    ],
                    [
                        gen_anomaly.sel(season="DJF").values,
                        gen_anomaly.sel(season="MAM").values,
                        gen_anomaly.sel(season="JJA").values,
                        gen_anomaly.sel(season="SON").values,
                    ],
                ],
                diverging=True,
                caption=(
                    f"Seasonal time-mean anomaly of {caption_name} for target (top) "
                    f"and gen (bottom) starting with DJF, R2={r2:.4f}. "
                    "Time-mean of target is subtracted from predictions and target."
                ),
            )
            plots[f"anomaly/{name}"] = image

            image_err = plot_paneled_data(
                [
                    [
                        bias[name].sel(season="DJF").values,
                        bias[name].sel(season="MAM").values,
                    ],
                    [
                        bias[name].sel(season="JJA").values,
                        bias[name].sel(season="SON").values,
                    ],
                ],
                diverging=True,
                caption=(
                    f"Seasonal bias of {caption_name} for DJF (Upper-Left), "
                    "MAM (UR), JJA (LL), and SON (LR). "
                    f"Seasonal anomaly R2={r2:.4f} (excludes time-mean of target)."
                ),
            )
            plots[f"bias/{name}"] = image_err

            mse_tensor = self._area_weighted_mean(
                torch.as_tensor(bias[name].values ** 2),
                name=name,
            )
            for i, season in enumerate(bias[name].season.values):
                rmse = float(mse_tensor[i].sqrt().numpy())
                metric_logs[f"time-mean-rmse/{name}-{season}"] = rmse
            rmse = float(
                # must compute area mean and then mean across seasons
                # before sqrt, so we can't use metrics.root_mean_squared_error
                mse_tensor.mean().sqrt().numpy()
            )
            metric_logs[f"time-mean-rmse/{name}"] = rmse

        if len(label) > 0:
            label = label + "/"
        logs: dict[str, Image | float] = {}
        logs.update({f"{label}{name}": plots[name] for name in plots.keys()})
        logs.update({f"{label}{name}": val for name, val in metric_logs.items()})
        return logs

    def get_dataset(self) -> xr.Dataset:
        logging.debug(
            "get_dataset not implemented for SeasonalAggregator. "
            "Returning an empty dataset."
        )
        return xr.Dataset()


ALL_SEASONS = np.asarray(["DJF", "MAM", "JJA", "SON"])


def _add_dataarray(da1: xr.DataArray, da2: xr.DataArray):
    """
    Perform dataarray addition, assuming any missing season indices
    have zero values.
    """
    if len(da1.season) < 4:
        da1 = da1.reindex(season=ALL_SEASONS, fill_value=0)
    if len(da2.season) < 4:
        da2 = da2.reindex(season=ALL_SEASONS, fill_value=0)
    return da1 + da2


def get_r2(da: xr.DataArray, target: xr.DataArray) -> float:
    """Compute the R2 value of the target compared to the reference."""
    SS_ref = np.sum((target.values - np.mean(target.values)) ** 2)
    SS_pred = np.sum((da - target).values ** 2)
    return float(1 - SS_pred / SS_ref)


def _reduce_datasets(dist: Distributed, dataset: xr.Dataset) -> xr.Dataset | None:
    """
    Add the dataset across all processes.

    Requires all dataset variables have the same shape.
    """
    # collect all data into one torch.Tensor for gathering, sort for determinism
    names = sorted(list(dataset.data_vars))
    # 'counts' must be present in the data, but we don't want to pack it with the others
    names.remove("counts")
    for name in names:
        if dataset[name].shape != dataset[names[0]].shape:
            raise ValueError(
                f"Variable {name} has shape {dataset[name].shape} "
                f"which is not equal to {dataset[names[0]].shape}"
            )
    tensor = torch.stack(
        [torch.as_tensor(dataset[name].values) for name in names],
        dim=0,
    ).to(get_device())
    reduced = dist.reduce_sum(tensor).cpu()
    reduced_counts = dist.reduce_sum(
        torch.as_tensor(dataset["counts"].values).to(get_device())
    ).cpu()
    dataset_out = xr.Dataset(
        {name: (["season", "lat", "lon"], reduced[i]) for i, name in enumerate(names)},
        coords=dataset.coords,
    )
    dataset_out["counts"] = xr.DataArray(reduced_counts, dims=["season"])
    return dataset_out


@torch.no_grad()
def _to_dataset(data: TensorMapping, time: xr.DataArray) -> xr.Dataset:
    """Convert a dictionary of data to an xarray dataset."""
    assert time.dims == ("sample", "time")  # must be consistent with this module
    data_vars = {}
    for name, tensor in data.items():
        data_vars[name] = (["sample", "time", "lat", "lon"], tensor)
    data_vars["counts"] = (["sample", "time"], np.ones(shape=time.shape))
    return xr.Dataset(data_vars, coords={"valid_time": time})



import pytest
import torch

from fme.core.mask_provider import MaskProvider, NullMaskProvider


def test_mask_provider_init_error():
    with pytest.raises(ValueError, match="var_1"):
        _ = MaskProvider(
            masks={
                "mask_0": torch.rand(2, 2),
                "var_1": torch.rand(2, 2),
            }
        )


def test_mask_provider_get_mask_tensor_for():
    mask_0 = torch.rand(2, 2)
    mask_1 = torch.rand(2, 2)
    mask_temp_1 = torch.rand(2, 2)
    mask_2d = torch.rand(2, 2)
    provider = MaskProvider(
        masks={
            "mask_0": mask_0,
            "mask_1": mask_1,
            "mask_temp_1": mask_temp_1,
            "mask_2d": mask_2d,
        }
    )
    # variable specific mask
    assert torch.equal(provider.get_mask_tensor_for("temp_1"), mask_temp_1)
    # level specific masks
    assert torch.equal(provider.get_mask_tensor_for("temp_0"), mask_0)
    assert torch.equal(provider.get_mask_tensor_for("humidity_0"), mask_0)
    assert torch.equal(provider.get_mask_tensor_for("humidity_1"), mask_1)
    # surface mask for non-3D vars
    assert torch.equal(provider.get_mask_tensor_for("surface_var"), mask_2d)


def test_mask_provider_get_mask_tensor_for_no_3d():
    mask_temp_1 = torch.rand(2, 2)
    mask_2d = torch.rand(2, 2)
    provider = MaskProvider(
        masks={
            "mask_temp_1": mask_temp_1,
            "mask_2d": mask_2d,
        }
    )
    # variable specific mask
    assert torch.equal(provider.get_mask_tensor_for("temp_1"), mask_temp_1)
    # 3D var doesn't trigger return of surface mask
    assert provider.get_mask_tensor_for("temp_0") is None
    # surface mask returned
    assert torch.equal(provider.get_mask_tensor_for("surface_var"), mask_2d)


def test_mask_provider_get_mask_tensor_for_var_specific_only():
    mask_temp_1 = torch.rand(2, 2)
    mask_var = torch.rand(2, 2)
    provider = MaskProvider(
        masks={
            "mask_temp_1": mask_temp_1,
            "mask_var": mask_var,
        }
    )
    # variable specific masks
    assert torch.equal(provider.get_mask_tensor_for("temp_1"), mask_temp_1)
    assert torch.equal(provider.get_mask_tensor_for("var"), mask_var)
    # other 3D vars have no mask
    assert provider.get_mask_tensor_for("temp_0") is None
    # other 2D vars have no mask
    assert provider.get_mask_tensor_for("var_2d") is None


def test_mask_provider_empty_init():
    provider = MaskProvider()
    assert provider.get_mask_tensor_for("temp_0") is None
    assert provider.get_mask_tensor_for("var") is None


def test_mask_provider_to_device():
    # test moving masks to a different device
    masks = {"mask_temp": torch.tensor([1, 0], dtype=torch.float32)}
    provider_cpu = MaskProvider(masks=masks)
    # using 'meta' device for testing without needing specific hardware
    provider_meta = provider_cpu.to("meta")
    assert provider_meta.masks["mask_temp"].device == torch.device("meta")
    # original provider should be unchanged
    assert provider_cpu.masks["mask_temp"].device == torch.device("cpu")


def test_mask_provider_update_success():
    # test updating masks from another provider with non-overlapping keys
    masks1 = {"mask_temp": torch.tensor([1, 0])}
    masks2 = {"mask_humidity": torch.tensor([0, 1])}
    provider1 = MaskProvider(masks=masks1)
    provider2 = MaskProvider(masks=masks2)

    provider1.update(provider2)

    expected_masks = {
        "mask_temp": torch.tensor([1, 0]),
        "mask_humidity": torch.tensor([0, 1]),
    }

    assert provider1.masks.keys() == expected_masks.keys()
    for name, mask in provider1.masks.items():
        assert torch.equal(mask, expected_masks[name])


def test_mask_provider_update_failure_overlapping_keys():
    # test updating masks fails when keys overlap
    masks1 = {"mask_temp": torch.tensor([1, 0]), "mask_common": torch.tensor([0, 0])}
    masks2 = {
        "mask_humidity": torch.tensor([0, 1]),
        "mask_common": torch.tensor([1, 1]),
    }
    provider1 = MaskProvider(masks=masks1)
    provider2 = MaskProvider(masks=masks2)

    with pytest.raises(ValueError, match="mask_common"):
        provider1.update(provider2)

    # Ensure original provider is unchanged
    assert provider1.masks.keys() == masks1.keys()
    for name, mask in provider1.masks.items():
        assert torch.equal(mask, masks1[name])


@pytest.mark.parametrize(
    "masks1, masks2, expected_equal",
    [
        pytest.param(
            {"mask_temp": torch.tensor([1, 0])},
            {"mask_temp": torch.tensor([1, 0])},
            True,
            id="providers_equal",
        ),
        pytest.param(
            {"mask_temp": torch.tensor([1, 0])},
            {"mask_temp": torch.tensor([1, 1])},
            False,
            id="providers_unequal",
        ),
        pytest.param(
            {"mask_temp": torch.tensor([1, 1]), "mask_humidity": torch.tensor([1, 1])},
            {"mask_temp": torch.tensor([1, 0])},
            False,
            id="first_provider_extra_key",
        ),
        pytest.param(
            {"mask_temp": torch.tensor([1, 0])},
            {"mask_temp": torch.tensor([1, 1]), "mask_humidity": torch.tensor([1, 1])},
            False,
            id="second_provider_extra_key",
        ),
        pytest.param(
            {"mask_temp": torch.ones(2, 2)},
            {"mask_temp": torch.ones(4, 4)},
            False,
            id="providers_mask_shapes_different",
        ),
    ],
)
def test_mask_provider_eq(masks1, masks2, expected_equal: bool):
    provider1 = MaskProvider(masks=masks1)
    provider2 = MaskProvider(masks=masks2)
    if expected_equal:
        assert provider1 == provider2
    else:
        assert provider1 != provider2


@pytest.mark.parametrize(
    "mask_provider",
    [
        pytest.param(MaskProvider(), id="empty"),
        pytest.param(
            MaskProvider(masks={"mask_0": torch.ones(10, 10)}), id="single_mask"
        ),
        pytest.param(
            MaskProvider(
                masks={
                    "mask_0": torch.ones(10, 10),
                    "mask_temp": torch.zeros(10, 10),
                }
            ),
            id="multiple_masks",
        ),
    ],
)
def test_mask_provider_round_trip(mask_provider: MaskProvider):
    state = mask_provider.to_state()
    reloaded_provider = MaskProvider.from_state(state)
    assert mask_provider == reloaded_provider


def test_null_mask_provider_update_err():
    mask_provider = MaskProvider(masks={"mask_2d": torch.rand(2, 2)})
    with pytest.raises(ValueError, match="mask_2d"):
        NullMaskProvider.update(mask_provider)



import datetime
import math

import pytest
import torch

from fme import get_device
from fme.core.coordinates import DepthCoordinate
from fme.core.corrector.ocean import (
    OceanCorrector,
    OceanCorrectorConfig,
    SeaIceFractionConfig,
)
from fme.core.gridded_ops import LatLonOperations
from fme.core.masking import StaticMaskingConfig
from fme.core.ocean_data import OceanData
from fme.core.typing_ import TensorMapping


def test_ocean_corrector_init_error():
    ops = LatLonOperations(torch.ones(size=IMG_SHAPE))
    timestep = datetime.timedelta(seconds=3600)

    config = OceanCorrectorConfig()
    # no error
    _ = OceanCorrector(config, ops, None, timestep)

    config = OceanCorrectorConfig(
        masking=StaticMaskingConfig(
            mask_value=0,
            fill_value=float("nan"),
        ),
    )
    with pytest.raises(
        ValueError,
        match="OceanCorrector.masking configured but DepthCoordinate missing.",
    ):
        _ = OceanCorrector(config, ops, None, timestep)


DEVICE = get_device()
IMG_SHAPE = (5, 5)
NZ = 2

_MASK = torch.ones(*IMG_SHAPE, NZ, device=DEVICE)
_LAT, _LON = 2, 2
_MASK[_LAT, _LON, :] = 0.0


class _MockDepth:
    def __len__(self) -> int:
        return len(self.get_idepth())

    def get_mask(self) -> torch.Tensor:
        return _MASK

    def get_mask_level(self, level: int):
        return _MASK.select(-1, level)

    def get_mask_tensor_for(self, name: str) -> torch.Tensor | None:
        if name.endswith("_1"):
            return _MASK.select(-1, 1)
        else:
            return _MASK.select(-1, 0)

    def get_idepth(self) -> torch.Tensor:
        return torch.tensor([0, 5, 15], device=DEVICE)

    def depth_integral(self, integrand: torch.Tensor) -> torch.Tensor:
        thickness = self.get_idepth().diff(dim=-1)
        return torch.nansum(_MASK * integrand * thickness, dim=-1)

    def build_output_masker(self):
        raise NotImplementedError

    def to(self, device: str) -> "_MockDepth":
        return self


_VERTICAL_COORD = _MockDepth()


@pytest.mark.parametrize("fill_value", [float("nan"), -1.0, 100.0])
def test_ocean_corrector_integration(fill_value):
    """Ensures that OceanCorrector can be called with all methods active
    but doesn't check results."""
    torch.manual_seed(0)
    config = OceanCorrectorConfig(
        masking=StaticMaskingConfig(
            mask_value=0,
            fill_value=fill_value,
        ),
        force_positive_names=["so_0", "so_1"],
    )
    ops = LatLonOperations(torch.ones(size=IMG_SHAPE))
    timestep = datetime.timedelta(seconds=3600)
    corrector = OceanCorrector(config, ops, _VERTICAL_COORD, timestep)
    input_data = {f"so_{i}": torch.randn(IMG_SHAPE, device=DEVICE) for i in range(NZ)}
    input_data["sst"] = torch.randn(IMG_SHAPE, device=DEVICE)
    gen_data = {f"so_{i}": torch.randn(IMG_SHAPE, device=DEVICE) for i in range(NZ)}
    gen_data["sst"] = torch.randn(IMG_SHAPE, device=DEVICE)
    corrected_gen = corrector(input_data, gen_data, {})
    for name in ["so_0", "so_1", "sst"]:
        if math.isnan(fill_value):
            assert corrected_gen[name][_LAT, _LON].isnan().item()
        else:
            torch.testing.assert_close(
                corrected_gen[name][_LAT, _LON],
                torch.tensor(fill_value, device=DEVICE),
                rtol=0,
                atol=0,
            )
    for name in ["so_0", "so_1"]:
        x = corrected_gen[name].clone()
        x[_LAT, _LON] = 0.0
        assert torch.all(x >= 0.0)


def test_ocean_corrector_has_no_negative_ocean_fraction():
    config = OceanCorrectorConfig(
        sea_ice_fraction_correction=SeaIceFractionConfig(
            sea_ice_fraction_name="sea_ice_fraction",
            land_fraction_name="land_fraction",
        ),
    )
    ops = LatLonOperations(torch.ones(size=IMG_SHAPE))
    timestep = datetime.timedelta(seconds=3600)
    input_data = {f"so_{i}": torch.randn(IMG_SHAPE, device=DEVICE) for i in range(NZ)}
    input_data["sst"] = torch.randn(IMG_SHAPE, device=DEVICE)
    input_data["land_fraction"] = torch.ones(IMG_SHAPE, device=DEVICE) * 0.8
    gen_data = {f"so_{i}": torch.randn(IMG_SHAPE, device=DEVICE) for i in range(NZ)}
    gen_data["sst"] = torch.randn(IMG_SHAPE, device=DEVICE)
    gen_data["sea_ice_fraction"] = torch.randn(IMG_SHAPE, device=DEVICE) * 0.5
    gen_data["sea_ice_fraction"][_LAT, _LON] = -0.5
    corrector = OceanCorrector(config, ops, None, timestep)
    violation = (input_data["land_fraction"] + gen_data["sea_ice_fraction"]) > 1.0
    assert violation.any()
    negative_sea_ice_fraction = gen_data["sea_ice_fraction"] < 0.0
    assert negative_sea_ice_fraction.any()

    next_step_input_data: TensorMapping = {}
    gen_data_corrected = corrector(input_data, gen_data, next_step_input_data)
    corrected_violation = (
        input_data["land_fraction"] + gen_data_corrected["sea_ice_fraction"]
    ) > 1.0
    assert not corrected_violation.any()
    assert not (gen_data_corrected["sea_ice_fraction"] < 0.0).any()


def test_ocean_corrector_has_negative_ocean_fraction():
    config = OceanCorrectorConfig(
        sea_ice_fraction_correction=SeaIceFractionConfig(
            sea_ice_fraction_name="sea_ice_fraction",
            land_fraction_name="land_fraction",
            remove_negative_ocean_fraction=False,
        ),
    )
    ops = LatLonOperations(torch.ones(size=IMG_SHAPE))
    timestep = datetime.timedelta(seconds=3600)
    input_data = {f"so_{i}": torch.randn(IMG_SHAPE, device=DEVICE) for i in range(NZ)}
    input_data["sst"] = torch.randn(IMG_SHAPE, device=DEVICE)
    input_data["land_fraction"] = torch.ones(IMG_SHAPE, device=DEVICE) * 0.8
    gen_data = {f"so_{i}": torch.randn(IMG_SHAPE, device=DEVICE) for i in range(NZ)}
    gen_data["sst"] = torch.randn(IMG_SHAPE, device=DEVICE)
    gen_data["sea_ice_fraction"] = torch.randn(IMG_SHAPE, device=DEVICE) * 0.5
    gen_data["sea_ice_fraction"][_LAT, _LON] = -0.5
    corrector = OceanCorrector(config, ops, None, timestep)
    violation = (input_data["land_fraction"] + gen_data["sea_ice_fraction"]) > 1.0
    assert violation.any()
    negative_sea_ice_fraction = gen_data["sea_ice_fraction"] < 0.0
    assert negative_sea_ice_fraction.any()

    next_step_input_data: TensorMapping = {}
    gen_data_corrected = corrector(input_data, gen_data, next_step_input_data)
    corrected_violation = (
        input_data["land_fraction"] + gen_data_corrected["sea_ice_fraction"]
    ) > 1.0
    assert corrected_violation.any()
    # sea_ice_fraction values are still clamped to [0, 1]
    assert not (gen_data_corrected["sea_ice_fraction"] < 0.0).any()


def test_sea_ice_thickness_correction():
    config = OceanCorrectorConfig(
        sea_ice_fraction_correction=SeaIceFractionConfig(
            sea_ice_fraction_name="sea_ice_fraction",
            land_fraction_name="land_fraction",
            sea_ice_thickness_name="HI",
        ),
    )
    ops = LatLonOperations(torch.ones(size=IMG_SHAPE))
    timestep = datetime.timedelta(seconds=3600)
    input_data = {"land_fraction": torch.ones(IMG_SHAPE, device=DEVICE)}
    input_data["land_fraction"][:3, :3] = torch.rand(3, 3, device=DEVICE)
    gen_data = {
        "sea_ice_fraction": torch.rand(IMG_SHAPE, device=DEVICE),
        "HI": torch.rand(IMG_SHAPE, device=DEVICE) * 10,
    }
    corrector = OceanCorrector(config, ops, None, timestep)
    gen_data_corrected = corrector(input_data, gen_data, {})
    sea_ice_zero = gen_data_corrected["sea_ice_fraction"] == 0.0
    thickness = gen_data_corrected["HI"]
    torch.testing.assert_close(
        torch.where(sea_ice_zero, thickness, 0.0), torch.zeros_like(thickness)
    )


def test_ocean_heat_content_correction():
    config = OceanCorrectorConfig(ocean_heat_content_correction=True)
    ops = LatLonOperations(torch.ones(size=[3, 3]))
    timestep = datetime.timedelta(seconds=5 * 24 * 3600)
    nsamples, nlat, nlon, nlevels = 4, 3, 3, 2
    mask = torch.ones(nsamples, nlat, nlon, nlevels)
    mask[:, 0, 0, 0] = 0.0
    mask[:, 0, 0, 1] = 0.0
    mask[:, 0, 1, 1] = 0.0
    idepth = torch.tensor([2.5, 10, 20])
    depth_coordinate = DepthCoordinate(idepth, mask)

    input_data_dict = {
        "thetao_0": torch.ones(nsamples, nlat, nlon),
        "thetao_1": torch.ones(nsamples, nlat, nlon),
        "hfds": torch.ones(nsamples, nlat, nlon),
        "hfgeou": torch.ones(nsamples, nlat, nlon),
        "sea_surface_fraction": mask[:, :, :, 0],
    }
    gen_data_dict = {
        "thetao_0": torch.ones(nsamples, nlat, nlon) * 2,
        "thetao_1": torch.ones(nsamples, nlat, nlon) * 2,
    }
    input_data = OceanData(input_data_dict, depth_coordinate)
    gen_data = OceanData(gen_data_dict, depth_coordinate)
    corrector = OceanCorrector(config, ops, depth_coordinate, timestep)
    gen_data_corrected_dict = corrector(input_data_dict, gen_data_dict, {})
    gen_data_corrected = OceanData(gen_data_corrected_dict, depth_coordinate)

    input_ohc = input_data.ocean_heat_content.sum(dim=(-1, -2), keepdim=True)
    gen_ohc = gen_data.ocean_heat_content.sum(dim=(-1, -2), keepdim=True)
    torch.allclose(
        gen_ohc,
        input_ohc * 2,
        atol=1e-10,
        equal_nan=True,
    )
    ohc_change = (
        2 * timestep.total_seconds() * 8
    )  # 2 because of hfds + hfgeou and 8 because of mask
    corrector_ratio = (input_ohc + ohc_change) / gen_ohc
    expected_gen_data_dict = {
        key: value * corrector_ratio for key, value in gen_data_dict.items()
    }
    expected_gen_data = OceanData(expected_gen_data_dict, depth_coordinate)
    torch.allclose(
        expected_gen_data.ocean_heat_content.nansum(dim=(-1, -2)),
        gen_data_corrected.ocean_heat_content.nansum(dim=(-1, -2)),
        atol=1e-10,
        equal_nan=True,
    )



import dataclasses
from collections.abc import Callable, Mapping
from typing import Any

import torch
from torch import nn

from fme.core.device import get_device
from fme.core.training_history import TrainingHistory
from fme.core.weight_ops import overwrite_weights
from fme.core.wildcard import apply_by_wildcard, wildcard_match

Weights = list[Mapping[str, Any]]
StepperWeightsAndHistory = tuple[Weights, TrainingHistory]


@dataclasses.dataclass
class FrozenParameterConfig:
    """
    Configuration for freezing parameters in a model.

    Parameter names can include wildcards, e.g. "encoder.*" will select
    all parameters in the encoder, while "encoder.*.bias" will select all
    bias parameters in the encoder. All parameters must be specified
    in either the include or exclude list, or
    an exception will be raised.

    An exception is raised if a parameter is included by both lists.

    Parameters:
        include: list of parameter names to freeze (set requires_grad = False)
        exclude: list of parameter names to ignore
    """

    include: list[str] = dataclasses.field(default_factory=list)
    exclude: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        for pattern in self.include:
            if any(wildcard_match(pattern, exclude) for exclude in self.exclude):
                raise ValueError(
                    f"Parameter {pattern} is included in both include "
                    f"{self.include} and exclude {self.exclude}"
                )
        for pattern in self.exclude:
            if any(wildcard_match(pattern, include) for include in self.include):
                raise ValueError(
                    f"Parameter {pattern} is included in both include "
                    f"{self.include} and exclude {self.exclude}"
                )

    def apply(self, model: nn.Module):
        apply_by_wildcard(model, _freeze_weight, self.include, self.exclude)


def _freeze_weight(module: nn.Module, name: str):
    try:
        module.get_parameter(name).requires_grad = False
    except AttributeError:  # non-parameter state
        pass


RegularizerFunction = Callable[[], torch.Tensor]


@dataclasses.dataclass
class ParameterClassification:
    """
    Specifies whether parameters are excluded from initialization or frozen.

    Parameters:
        exclude_parameters: list of parameter names to exclude from the loaded
            weights. Used for example to keep the random initialization for
            final layer(s) of a model, and only overwrite the weights for
            earlier layers. Takes values like "decoder.2.weight".
        frozen_parameters: configuration for freezing parameters in the built model
    """

    exclude: list[str] = dataclasses.field(default_factory=list)
    frozen: FrozenParameterConfig = dataclasses.field(
        default_factory=lambda: FrozenParameterConfig(exclude=["*"])
    )


@dataclasses.dataclass
class ParameterInitializationConfig:
    """
    A class which applies custom initialization to module parameters.

    Assumes the module weights have already been randomly initialized.

    Supports overwriting the weights of the built model with weights from a
    pre-trained model. If the built model has larger weights than the
    pre-trained model, only the initial slice of the weights is overwritten.

    Parameters:
        weight_path: path to a Stepper checkpoint
            containing weights to load
        parameters: list of ParameterClassification objects, each specifying
            whether parameters are excluded from initialization or frozen.
            By default modules are unfrozen and all parameters are included.
            Must be provided in the same order as provided by the stepper's
            `.modules` attribute.
        alpha: L2 regularization coefficient keeping initialized weights
            close to their intiial values
        beta: L2 regularization coefficient keeping uninitialized weights
            close to zero
        exclude_parameters: deprecated, kept for backwards compatibility
        frozen_parameters: deprecated, kept for backwards compatibility
    """

    weights_path: str | None = None
    parameters: list[ParameterClassification] = dataclasses.field(default_factory=list)
    alpha: float = 0.0
    beta: float = 0.0
    exclude_parameters: list[str] | None = None
    frozen_parameters: FrozenParameterConfig | None = None

    def __post_init__(self):
        if self.exclude_parameters is not None or self.frozen_parameters is not None:
            if len(self.parameters) > 0:
                raise ValueError(
                    "exclude_parameters and frozen_parameters are deprecated, "
                    "do not provide both parameters and exclude_parameters or "
                    "frozen_parameters"
                )
            exclude = self.exclude_parameters or []
            frozen = self.frozen_parameters or FrozenParameterConfig(exclude=["*"])
            self.parameters = [ParameterClassification(exclude=exclude, frozen=frozen)]
        self.exclude_parameters = None
        self.frozen_parameters = None

    def build(
        self,
        load_weights_and_history: Callable[[str], StepperWeightsAndHistory],
    ) -> "ParameterInitializer":
        """
        Build a ParameterInitializer instance with the current configuration.

        Args:
            load_weights_and_history: a function which loads weights and training
                history from a path, specifically the configured weights_path.
        """
        return ParameterInitializer(
            config=self, load_weights_and_history=load_weights_and_history
        )


@dataclasses.dataclass
class ParameterInitializer:
    config: ParameterInitializationConfig = dataclasses.field(
        default_factory=ParameterInitializationConfig
    )
    load_weights_and_history: Callable[[str], StepperWeightsAndHistory] = (
        dataclasses.field(default=lambda _: ([], TrainingHistory()))
    )

    def __post_init__(self):
        self._base_weights: Weights | None = None
        self._training_history: TrainingHistory | None = None

    @property
    def base_weights(self) -> Weights | None:
        if self.config.weights_path is not None and self._base_weights is None:
            self._base_weights, self._training_history = self.load_weights_and_history(
                self.config.weights_path
            )
        return self._base_weights

    @property
    def training_history(self) -> TrainingHistory | None:
        if self.config.weights_path is not None and self._training_history is None:
            self._base_weights, self._training_history = self.load_weights_and_history(
                self.config.weights_path
            )
        return self._training_history

    def _filled_parameters(self, n_modules: int) -> list[ParameterClassification]:
        return self.config.parameters + [
            ParameterClassification()
            for _ in range(n_modules - len(self.config.parameters))
        ]

    def apply_weights(
        self,
        modules: list[nn.Module],
    ) -> None:
        """
        Apply the weight initialization from a base model to a module.

        Args:
            modules: a list of nn.Modules to initialize
        """
        filled_parameters = self._filled_parameters(len(modules))
        if self.config.weights_path is not None:
            if self.base_weights is not None:
                for module, state_dict, classification in zip(
                    modules, self.base_weights, filled_parameters
                ):
                    overwrite_weights(
                        state_dict,
                        module,
                        exclude_parameters=classification.exclude,
                    )

    def freeze_weights(self, modules: list[nn.Module]):
        """
        Freeze the weights of the modules.

        Note this must be called before wrapping the modules in a DDP layer,
        otherwise the DistributedDataParallel will expect frozen
        weights to have gradients, and will raise an exception.

        Args:
            modules: a list of nn.Modules to freeze if configured to do so.
        """
        filled_parameters = self._filled_parameters(len(modules))
        for module, classification in zip(modules, filled_parameters):
            classification.frozen.apply(module)

    def get_l2_sp_tuning_regularizer(
        self,
        modules: list[nn.Module],
    ) -> RegularizerFunction:
        """Get L2-SP loss regularizer function for the parameters of the modules.
        The regularizer function computes the L2 regularization loss based on
        the base weights and the current weights of the modules.

        If the base weights are set, it computes the L2 regularization loss
        for each module based on the difference between the current weights
        and the base weights, and the L2 norm of the weights that are not
        initialized (i.e., those that are excluded from initialization).

        Args:
            modules: a list of nn.Modules to compute the regularization for

        Returns:
            A function that returns a tensor representing the regularization loss.
        """
        device = get_device()
        filled_parameters = self._filled_parameters(len(modules))
        base_weights = self.base_weights
        if base_weights is not None and (
            self.config.alpha != 0 or self.config.beta != 0
        ):
            for module, state_dict in zip(modules, base_weights):
                state_dict = {
                    name: value.to(device) for name, value in state_dict.items()
                }
                from_names = set(state_dict.keys())
                to_names = set(module.state_dict().keys())
                if not from_names.issubset(to_names):
                    missing_parameters = from_names - to_names
                    raise ValueError(
                        f"Dest module is missing parameters {missing_parameters}, "
                        "which is not allowed"
                    )

            def regularizer():
                loss = torch.tensor(0.0, device=device)
                for module, state_dict, classification in zip(
                    modules, base_weights, filled_parameters
                ):
                    for name in state_dict.keys():
                        try:
                            param = module.get_parameter(name)
                        except AttributeError:  # non-trainable state data
                            continue
                        if any(
                            wildcard_match(pattern, name)
                            for pattern in classification.exclude
                        ):
                            loss += (
                                self.config.beta
                                / 2
                                * torch.linalg.norm(param.flatten(), ord=2)
                            )
                        else:
                            loss += (
                                self.config.alpha
                                / 2
                                * torch.linalg.norm(
                                    (param - state_dict[name]).flatten(),
                                    ord=2,
                                )
                            )

                return loss

        else:

            def regularizer():
                return torch.tensor(0.0, device=device)

            return regularizer

        return regularizer



import time

import matplotlib.figure
import numpy as np
import pytest
import torch

from fme.core.histogram import (
    ComparedDynamicHistograms,
    DynamicHistogram,
    _normalize_histogram,
)


def test__normalize_histogram():
    counts = np.array(
        [
            1,
            2,
            3,
            4,
        ]
    )
    bin_edges = np.array([0, 1, 2, 3, 5])
    normalized_counts = _normalize_histogram(counts, bin_edges)
    assert sum(np.diff(bin_edges) * normalized_counts) == 1.0


@pytest.mark.parametrize(
    "n_times, time_bin_len",
    [
        pytest.param(3, 3, id="one_time_bin"),
        pytest.param(6, 2, id="multiple_time_bins"),
    ],
)
def test_dynamic_histogram_random_values(n_times: int, time_bin_len: int):
    np.random.seed(0)
    n_values = 50_000
    n_window = 100
    # gradually increasing values in time ensure bins get expanded
    data = (
        np.random.uniform(low=0, high=1, size=(n_times, n_values))
        * np.arange(1, n_values + 1)[None, :]
    )
    histogram = DynamicHistogram(n_times=n_times, n_bins=20)
    start = time.time()
    for i_time in range(0, n_times, time_bin_len):
        for i in range(0, n_values, n_window):
            histogram.add(
                torch.as_tensor(data[i_time : i_time + time_bin_len, i : i + n_window]),
                i_time_start=i_time,
            )
    end = time.time()
    assert (end - start) < 0.8, "histogram too slow"
    assert np.sum(histogram.counts) == n_times * n_values
    direct_histogram = np.apply_along_axis(
        lambda arr: np.histogram(arr, bins=histogram.bin_edges)[0],
        axis=1,
        arr=data,
    )
    assert np.allclose(histogram.counts, direct_histogram)
    assert histogram.bin_edges is not None
    histogram_range = histogram.bin_edges[-1] - histogram.bin_edges[0]
    value_range = np.max(data) - np.min(data)

    # values may have been doubled to the left and then to the right,
    # and if the values were infinitesmally to the left and right of
    # the existing range when that happens, there will be 3 "empty"
    # sections to the one original "full" section:
    #     [ 0, #, 0, 0]
    # where there could be a value on the rightmost edge of the first
    # and leftmost edge of the third bin, with none outside that range.
    #
    # There should be no way to get a histogram range larger than this,
    # relative to the value range.
    assert histogram_range <= 4.0 * value_range


def test_dynamic_histogram_extends_as_expected():
    histogram = DynamicHistogram(n_times=1, n_bins=200)
    histogram.add(torch.as_tensor([[-1.0, 0.0, 1.0]]))
    bin_edges = histogram.bin_edges
    assert bin_edges is not None
    np.testing.assert_approx_equal(bin_edges[0], -1.0, significant=6)
    np.testing.assert_approx_equal(bin_edges[-1], 1.0, significant=6)
    histogram.add(torch.as_tensor([[-2.0]]))
    bin_edges = histogram.bin_edges
    assert bin_edges is not None
    # double in size to the left, length becomes 4, from -3 to 1.0
    np.testing.assert_approx_equal(bin_edges[0], -3.0, significant=6)
    np.testing.assert_approx_equal(bin_edges[-1], 1.0, significant=6)
    histogram.add(torch.as_tensor([[2.0]]))
    bin_edges = histogram.bin_edges
    assert bin_edges is not None
    # double in size to the right, length becomes 8, from -3 to 5.0
    np.testing.assert_approx_equal(bin_edges[0], -3.0, significant=6)
    np.testing.assert_approx_equal(bin_edges[-1], 5.0, significant=6)
    histogram.add(torch.as_tensor([[27.0]]))
    bin_edges = histogram.bin_edges
    assert bin_edges is not None
    # double in size twice to the right, length becomes 32, from -3 to 29.0
    np.testing.assert_approx_equal(bin_edges[0], -3.0, significant=6)
    np.testing.assert_approx_equal(bin_edges[-1], 29.0, significant=6)


def test_histogram_handles_uniform_field():
    histogram = DynamicHistogram(n_times=1, n_bins=200)
    histogram.add(torch.as_tensor([[1.0, 1.0, 1.0]]))  # has zero range
    histogram.add(torch.as_tensor([[1.0, 2.0, 3.0]]))  # has non-zero range


@pytest.mark.parametrize(
    "shape",
    [
        pytest.param((2, 8, 16), id="no_time_dim"),
        pytest.param((2, 1, 8, 16), id="time_dim"),
    ],
)
@pytest.mark.parametrize("percentiles", [[], [99.0], [99.0, 99.99]])
def test_compared_dynamic_histograms(shape, percentiles):
    n_bins = 300
    histogram = ComparedDynamicHistograms(n_bins, percentiles=percentiles)
    target = {"x": torch.ones(*shape), "y": torch.zeros(*shape)}
    prediction = {"x": torch.rand(*shape), "y": torch.rand(*shape)}
    histogram.record_batch(target, prediction)
    wandb_result = histogram.get_wandb()

    percentile_names = []
    for p in percentiles:
        for data_type in ("target", "prediction"):
            for var_name in ("x", "y"):
                percentile_names.append(f"{data_type}/{p}th-percentile/{var_name}")

    assert sorted(list(wandb_result.keys())) == sorted(
        [
            "x",
            "y",
        ]
        + percentile_names
    )
    for var_name in ["x", "y"]:
        for data_type in ["target", "prediction"]:
            assert isinstance(wandb_result[f"{var_name}"], matplotlib.figure.Figure)
            for p in percentiles:
                assert isinstance(
                    wandb_result[f"{data_type}/{p}th-percentile/{var_name}"], float
                )

    ds = histogram.get_dataset()
    all(ds.coords["source"] == ["target", "prediction"])


@pytest.mark.parametrize(
    "target, prediction",
    [
        pytest.param(
            {"x": torch.ones(2, 8, 16), "y": torch.zeros(2, 8, 16)},
            {},
            id="non_empty_target_empty_prediction",
        ),
        pytest.param(
            {},
            {"x": torch.rand(2, 8, 16), "y": torch.rand(2, 8, 16)},
            id="empty_target_non_empty_prediction",
        ),
        pytest.param({}, {}, id="empty_target_empty_prediction"),
    ],
)
def test_compared_dynamic_histogram_empty_tensor_mapping(target, prediction):
    """
    Test that histogram aggregator handles empty tensor mappings.
    """
    n_bins = 300
    histogram = ComparedDynamicHistograms(n_bins, percentiles=[99.0])

    with pytest.raises(ValueError):
        histogram.record_batch(target, prediction)


def test_compared_dynamic_histogram_varying_variables_with_record_batch():
    n_bins = 300
    histogram = ComparedDynamicHistograms(n_bins, percentiles=[99.0])
    target = {"x": torch.ones(2, 8, 16), "y": torch.zeros(2, 8, 16)}
    prediction = {"x": torch.rand(2, 8, 16), "y": torch.rand(2, 8, 16)}

    # record one batch to set the recorded variables
    histogram.record_batch(target, prediction)

    target_missing_y = {"x": torch.ones(2, 8, 16)}
    prediction_missing_y = {"x": torch.rand(2, 8, 16)}
    target_extra_z = {**target, "z": torch.ones(2, 8, 16)}
    prediction_extra_w = {**prediction, "w": torch.rand(2, 8, 16)}

    # fails on variables missing from inital recorded batch
    with pytest.raises(ValueError):
        histogram.record_batch(target_missing_y, prediction)
    with pytest.raises(ValueError):
        histogram.record_batch(target, prediction_missing_y)

    # ignores non-matching variables
    histogram.record_batch(target_extra_z, prediction_extra_w)


def test_compared_dynamic_histogram_with_nans_record_batch():
    n_bins = 300
    histogram = ComparedDynamicHistograms(n_bins, percentiles=[99.0])

    # record one batch
    target = {"x": torch.ones(2, 2, 8, 16), "y": torch.zeros(2, 2, 8, 16)}
    target["x"][0, 0, 0] = float("nan")
    prediction = {"x": torch.rand(2, 2, 8, 16), "y": torch.rand(2, 2, 8, 16)}
    prediction["x"][0, 0, 0] = float("nan")

    histogram.record_batch(target, prediction)
    wandb_result = histogram.get_wandb()
    assert not np.isnan(wandb_result["target/99.0th-percentile/x"])
    assert not np.isnan(wandb_result["prediction/99.0th-percentile/x"])

    # record another batch with different time dimension size
    target = {"x": torch.ones(2, 1, 8, 16), "y": torch.zeros(2, 1, 8, 16)}
    target["x"][0, 0, 0] = float("nan")
    prediction = {"x": torch.rand(2, 1, 8, 16), "y": torch.rand(2, 1, 8, 16)}
    prediction["x"][0, 0, 0] = float("nan")

    histogram.record_batch(target, prediction)
    wandb_result = histogram.get_wandb()
    assert not np.isnan(wandb_result["target/99.0th-percentile/x"])
    assert not np.isnan(wandb_result["prediction/99.0th-percentile/x"])



"""
This file contains code for computing metrics of single variables on batches of data,
and aggregating them into a single metric value. The functions here mainly exist
to turn metric functions that may have different APIs into a common API,
so that they can be iterated over and called in the same way in a loop.
"""

from collections import defaultdict
from typing import Protocol

import torch

from fme.core.typing_ import TensorDict, TensorMapping


class AreaWeightedFunction(Protocol):
    """
    A function that computes a metric on the true and predicted values,
    weighted by area.
    """

    def __call__(
        self,
        truth: TensorMapping,
        predicted: TensorMapping,
    ) -> TensorDict: ...


class ReducedMetric(Protocol):
    """Used to record a metric value on batches of data (potentially out-of-memory)
    and then get the total metric at the end.
    """

    def record(self, target: TensorMapping, gen: TensorMapping):
        """
        Update metric for a batch of data.
        """
        ...

    def get(self) -> TensorDict:
        """
        Get the total metric value, not divided by number of recorded batches.
        """
        ...


class AreaWeightedReducedMetric:
    """
    A wrapper around an area-weighted metric function.
    """

    def __init__(
        self,
        device: torch.device,
        compute_metric: AreaWeightedFunction,
    ):
        self._compute_metric = compute_metric
        self._total: TensorDict | None = None
        self._device = device

    def _update_total(self, tensors: TensorDict):
        if self._total is None:
            self._total = {
                name: torch.zeros_like(tensor, device=self._device)
                for name, tensor in tensors.items()
            }
        missing_names = set(self._total) - set(tensors)
        if len(missing_names) > 0:
            raise ValueError(
                f"Missing metrics for {missing_names} which were "
                "present the first time metrics were recorded."
            )
        for name, tensor in tensors.items():
            try:
                self._total[name] += tensor
            except KeyError:
                raise ValueError(
                    "Attempted to record the area weighted reduced metric for "
                    f"'{name}' but it was not present the first time metrics "
                    "were recorded."
                )

    def record(self, target: TensorMapping, gen: TensorMapping):
        """Add a batch of data to the metric.

        Args:
            target: Target data. Should have shape [batch, time, height, width].
            gen: Generated data. Should have shape [batch, time, height, width].
        """
        # Update totals for each variable
        batch_avgs = {
            name: tensor.mean(dim=0)
            for name, tensor in self._compute_metric(target, gen).items()
        }
        self._update_total(batch_avgs)

    def get(self) -> TensorDict:
        """Returns the metric."""
        if self._total is None:
            return defaultdict(lambda: torch.tensor(torch.nan))
        return self._total



from collections.abc import Iterable
from typing import TypeAlias

import numpy as np
import torch
import torch_harmonics

from fme.core.constants import GRAVITY, LATENT_HEAT_OF_FREEZING

Dimension: TypeAlias = int | Iterable[int]
Array: TypeAlias = np.ndarray | torch.Tensor


def spherical_area_weights(lats: Array, num_lon: int) -> torch.Tensor:
    """Computes area weights given the latitudes of a regular lat-lon grid.

    Args:
        lats: tensor of shape (..., num_lat,) with the latitudes of
            the cell centers.
        num_lon: Number of longitude points.

    Returns:
        a torch.tensor of shape (num_lat, num_lon).
    """
    if isinstance(lats, np.ndarray):
        lats = torch.from_numpy(lats)
    expand_shape = [-1 for _ in range(lats.dim())]
    weights = (
        torch.cos(torch.deg2rad(lats)).unsqueeze(-1).expand(expand_shape + [num_lon])
    )
    weights = weights / weights.sum(dim=(-1, -2), keepdim=True)
    return weights


def weighted_sum(
    tensor: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
    keepdim: bool = False,
) -> torch.Tensor:
    """Computes the weighted sum across the specified list of dimensions.

    Args:
        tensor: torch.Tensor
        weights: Weights to apply to the sum.
        dim: Dimensions to compute the sum over.
        keepdim: Whether the output tensor has `dim` retained or not.

    Returns:
        a tensor of the weighted sum averaged over the specified dimensions `dim`.
    """
    if weights is None:
        return tensor.sum(dim=dim, keepdim=keepdim)

    expanded_weights = weights.expand(tensor.shape)

    # remove potential "expected NaNs", i.e. any NaNs with 0 weight
    tensor = tensor.where(expanded_weights != 0.0, 0.0)

    return (tensor * expanded_weights).sum(dim=dim, keepdim=keepdim)


def weighted_mean(
    tensor: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
    keepdim: bool = False,
) -> torch.Tensor:
    """Computes the weighted mean across the specified list of dimensions.

    Args:
        tensor: torch.Tensor
        weights: Weights to apply to the mean.
        dim: Dimensions to compute the mean over.
        keepdim: Whether the output tensor has `dim` retained or not.

    Returns:
        a tensor of the weighted mean averaged over the specified dimensions `dim`.
    """
    if weights is None:
        return tensor.mean(dim=dim, keepdim=keepdim)

    expanded_weights = weights.expand(tensor.shape)

    # remove potential "expected NaNs", i.e. any NaNs with 0 weight
    tensor = tensor.where(expanded_weights != 0.0, 0.0)

    return (tensor * expanded_weights).sum(
        dim=dim, keepdim=keepdim
    ) / expanded_weights.sum(dim=dim, keepdim=keepdim)


def weighted_nanmean(
    tensor: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
    keepdim: bool = False,
) -> torch.Tensor:
    """Computes the weighted nanmean across the specified list of dimensions.

    Args:
        tensor: torch.Tensor
        weights: Weights to apply to the mean.
        dim: Dimensions to compute the mean over.
        keepdim: Whether the output tensor has `dim` retained or not.

    Returns:
        a tensor of the weighted mean averaged over the specified dimensions `dim`.
    """
    if weights is None:
        return tensor.nanmean(dim=dim, keepdim=keepdim)
    denom = torch.where(torch.isnan(tensor), 0.0, weights.expand(tensor.shape)).sum(
        dim=dim, keepdim=keepdim
    )
    return (tensor * weights).nansum(dim=dim, keepdim=keepdim) / denom


def weighted_std(
    tensor: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
) -> torch.Tensor:
    """Computes the weighted standard deviation across the specified list of dimensions.

    Computed by first computing the weighted variance, then taking the square root.

    weighted_variance = weighted_mean((tensor - weighted_mean(tensor)) ** 2)) ** 0.5

    Args:
        tensor: torch.Tensor
        weights: Weights to apply to the variance.
        dim: Dimensions to compute the standard deviation over.

    Returns:
        a tensor of the weighted standard deviation over the
            specified dimensions `dim`.
    """
    if weights is None:
        weights = torch.tensor(1.0, device=tensor.device)

    mean = weighted_mean(tensor, weights=weights, dim=dim, keepdim=True)
    variance = weighted_mean((tensor - mean) ** 2, weights=weights, dim=dim)
    return torch.sqrt(variance)


def weighted_mean_bias(
    truth: torch.Tensor,
    predicted: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
) -> torch.Tensor:
    """Computes the mean bias across the specified list of dimensions assuming
    that the weights are applied to the last dimensions, e.g. the spatial dimensions.

    Args:
        truth: torch.Tensor
        predicted: torch.Tensor
        dim: Dimensions to compute the mean over.
        weights: Weights to apply to the mean.

    Returns:
        a tensor of the mean biases averaged over the specified dimensions `dim`.
    """
    assert (
        truth.shape == predicted.shape
    ), "Truth and predicted should have the same shape."
    bias = predicted - truth
    return weighted_mean(bias, weights=weights, dim=dim)


def root_mean_squared_error(
    truth: torch.Tensor,
    predicted: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
) -> torch.Tensor:
    """
    Compute a weighted root mean square error between truth and predicted.

    Namely:

        sqrt((weights * ((xhat - x) ** 2)).mean(dims))

    Args:
        truth: torch.Tensor whose last dimensions are to be weighted
        predicted: torch.Tensor whose last dimensions are to be weighted
        weights: torch.Tensor to apply to the squared bias.
        dim: Dimensions to average over.

    Returns:
        A tensor of weighted RMSEs.
    """
    assert (
        truth.shape == predicted.shape
    ), "Truth and predicted should have the same shape."
    sq_bias = torch.square(predicted - truth)
    return weighted_mean(sq_bias, weights=weights, dim=dim).sqrt()


def gradient_magnitude(tensor: torch.Tensor, dim: Dimension = ()) -> torch.Tensor:
    """Compute the magnitude of gradient across the specified dimensions."""
    gradients = torch.gradient(tensor, dim=dim)
    return torch.sqrt(sum([g**2 for g in gradients]))


def weighted_mean_gradient_magnitude(
    tensor: torch.Tensor, weights: torch.Tensor | None = None, dim: Dimension = ()
) -> torch.Tensor:
    """Compute weighted mean of gradient magnitude across the specified dimensions."""
    return weighted_nanmean(gradient_magnitude(tensor, dim), weights=weights, dim=dim)


def gradient_magnitude_percent_diff(
    truth: torch.Tensor,
    predicted: torch.Tensor,
    weights: torch.Tensor | None = None,
    dim: Dimension = (),
) -> torch.Tensor:
    """Compute the percent difference of the weighted mean gradient magnitude across
    the specified dimensions.
    """
    truth_grad_mag = weighted_mean_gradient_magnitude(truth, weights, dim)
    predicted_grad_mag = weighted_mean_gradient_magnitude(predicted, weights, dim)
    return 100 * (predicted_grad_mag - truth_grad_mag) / truth_grad_mag


def rmse_of_time_mean(
    truth: torch.Tensor,
    predicted: torch.Tensor,
    weights: torch.Tensor | None = None,
    time_dim: Dimension = 0,
    spatial_dims: Dimension = (-2, -1),
) -> torch.Tensor:
    """Compute the RMSE of the time-average given truth and predicted.

    Args:
        truth: truth tensor
        predicted: predicted tensor
        weights: weights to use for computing spatial RMSE
        time_dim: time dimension
        spatial_dims: spatial dimensions over which RMSE is calculated

    Returns:
        The RMSE between the time-mean of the two input tensors. The time and
            spatial dims are reduced.
    """
    truth_time_mean = truth.mean(dim=time_dim)
    predicted_time_mean = predicted.mean(dim=time_dim)
    ret = root_mean_squared_error(
        truth_time_mean, predicted_time_mean, weights=weights, dim=spatial_dims
    )
    return ret


def time_and_global_mean_bias(
    truth: torch.Tensor,
    predicted: torch.Tensor,
    weights: torch.Tensor | None = None,
    time_dim: Dimension = 0,
    spatial_dims: Dimension = (-2, -1),
) -> torch.Tensor:
    """Compute the global- and time-mean bias given truth and predicted.

    Args:
        truth: truth tensor
        predicted: predicted tensor
        weights: weights to use for computing the global mean
        time_dim: time dimension
        spatial_dims: spatial dimensions over which global mean is calculated

    Returns:
        The global- and time-mean bias between the predicted and truth tensors. The
            time and spatial dims are reduced.
    """
    truth_time_mean = truth.mean(dim=time_dim)
    predicted_time_mean = predicted.mean(dim=time_dim)
    result = weighted_mean(
        predicted_time_mean - truth_time_mean, weights=weights, dim=spatial_dims
    )
    return result


def surface_pressure_due_to_dry_air(
    surface_pressure: torch.Tensor,
    total_water_path: torch.Tensor,
) -> torch.Tensor:
    """Computes the dry air (Pa).

    Args:
        surface_pressure: the surface preessure in Pa.
        total_water_path: the total water path in kg/m^2.

    Returns:
        The surface pressure due to dry air mass only. (Pa)
    """
    return surface_pressure - GRAVITY * total_water_path


def net_surface_energy_flux(
    lw_rad_down,
    lw_rad_up,
    sw_rad_down,
    sw_rad_up,
    latent_heat_flux,
    sensible_heat_flux,
    frozen_precipitation_rate=None,
) -> torch.Tensor:
    """
    Compute the net surface energy flux from individual terms in budget.

    Args:
        lw_rad_down: Downward longwave surface radiation in W/m^2.
        lw_rad_up: Upward longwave surface radiation in W/m^2.
        sw_rad_down: Downward shortwave surface radiation in W/m^2.
        sw_rad_up: Upward shortwave surface radiation in W/m^2.
        latent_heat_flux: Latent heat flux in W/m^2 (positive=upward).
        sensible_heat_flux: Sensible heat flux in W/m^2 (positive=upward).
        frozen_precipitation_rate (optional): Frozen precipitation rate in kg/m^2/s.

    Returns:
        Net surface energy flux in W/m^2. Positive values indicate energy flowing
        from atmosphere to surface.
    """
    if frozen_precipitation_rate is not None:
        frozen_precip_heat_flux = frozen_precipitation_rate * LATENT_HEAT_OF_FREEZING
    else:
        frozen_precip_heat_flux = 0.0
    net_surface_radiative_flux = sw_rad_down - sw_rad_up + lw_rad_down - lw_rad_up
    net_surface_turbulent_heat_flux = -latent_heat_flux - sensible_heat_flux
    return (
        net_surface_radiative_flux
        + net_surface_turbulent_heat_flux
        - frozen_precip_heat_flux
    )


def net_top_of_atmosphere_energy_flux(
    sw_rad_down, sw_rad_up, lw_rad_up
) -> torch.Tensor:
    """
    Compute the net top-of-atmosphere energy flux from individual terms in budget.

    Args:
        sw_rad_down: Downward shortwave TOA radiation in W/m^2.
        sw_rad_up: Upward longwave TOA radiation in W/m^2.
        lw_rad_up: Upward shortwave TOA radiation in W/m^2.

    Returns:
        Net TOA energy flux in W/m^2. Positive values indicate energy flowing
        into the atmosphere.
    """
    return sw_rad_down - sw_rad_up - lw_rad_up


def quantile(bins: np.ndarray, hist: np.ndarray, probability: float) -> float:
    """
    Calculate the quantile value for a given histogram, using linear
    interpolation within bins. This is the inverse of the cumulative
    distribution function (CDF).

    Args:
        bins: The bin edges of the histogram.
        hist: The histogram values.
        probability: The probability to query the quantile for.

    Returns:
        The quantile value.

    Raises:
        ValueError: If probability is not between 0 and 1.
    """
    if not (0.0 <= probability <= 1.0):
        raise ValueError("Probabilities must be between 0 and 1.")

    # get the normalized CDF based on the histogram
    cdf = np.cumsum(hist)
    cdf = cdf / cdf[-1]
    # append initial zero to cdf as there are no values less than the first bin
    cdf = np.insert(cdf, 0, 0)
    # find within which bin the requested pct percentile falls
    bin_idx = np.argmax(cdf > probability) - 1
    # linearly interpolate within the bin to get the percentile value
    c0, c1 = bins[bin_idx], bins[bin_idx + 1]
    p0, p1 = cdf[bin_idx], cdf[bin_idx + 1]
    return c0 + (c1 - c0) * (probability - p0) / (p1 - p0)


def spherical_power_spectrum(
    field: torch.Tensor, sht: torch_harmonics.RealSHT
) -> torch.Tensor:
    """Compute the spherical power spectrum of a field.

    Args:
        field: The field to compute the power spectrum for. It is assumed that
            the last two dimension are latitude and longitude, respectively.
        sht: An initialized spherical harmonics transformer. Must be passed for
            performance reasons.

    Returns:
        The power spectrum of the field. Will have one fewer dimensions than the
            input field.

    Notes:
        Computed by summing over all "m" wavenumbers for each total "l" wavenumber.
    """
    field_sht = sht.forward(field)
    power_spectrum = torch.sum(abs(field_sht) ** 2, dim=-1)
    return power_spectrum



import copy
import datetime
from collections.abc import Iterable, Mapping, Sequence
from pathlib import Path

import cftime
import numpy as np
import numpy.typing as npt
import torch
import xarray as xr
from netCDF4 import Dataset

from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.utils import (
    DIM_INFO_HEALPIX,
    DIM_INFO_LATLON,
    get_all_names,
)
from fme.core.dataset.data_typing import VariableMetadata

LEAD_TIME_DIM = "time"
LEAD_TIME_UNITS = "microseconds"
IC_DIM = "sample"
INIT_TIME = "init_time"
INIT_TIME_UNITS = "microseconds since 1970-01-01 00:00:00"
VALID_TIME = "valid_time"


class PairedRawDataWriter:
    """
    Wrapper over RawDataWriter to write both target and prediction data.

    Gives the same interface as for our other writers.
    """

    def __init__(
        self,
        path: str,
        n_initial_conditions: int,
        save_names: Sequence[str] | None,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
    ):
        self._target_writer = RawDataWriter(
            path=path,
            label="autoregressive_target.nc",
            n_initial_conditions=n_initial_conditions,
            save_names=save_names,
            variable_metadata=variable_metadata,
            coords=coords,
            dataset_metadata=dataset_metadata,
        )
        self._prediction_writer = RawDataWriter(
            path=path,
            label="autoregressive_predictions.nc",
            n_initial_conditions=n_initial_conditions,
            save_names=save_names,
            variable_metadata=variable_metadata,
            coords=coords,
            dataset_metadata=dataset_metadata,
        )

    def append_batch(
        self,
        target: dict[str, torch.Tensor],
        prediction: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        self._target_writer.append_batch(
            data=target,
            start_timestep=start_timestep,
            batch_time=batch_time,
        )
        self._prediction_writer.append_batch(
            data=prediction,
            start_timestep=start_timestep,
            batch_time=batch_time,
        )

    def flush(self):
        self._target_writer.flush()
        self._prediction_writer.flush()


class RawDataWriter:
    """
    Write raw data to a netCDF file.
    """

    def __init__(
        self,
        path: str,
        label: str,
        n_initial_conditions: int,
        save_names: Sequence[str] | None,
        variable_metadata: Mapping[str, VariableMetadata],
        coords: Mapping[str, np.ndarray],
        dataset_metadata: DatasetMetadata,
    ):
        """
        Args:
            path: Directory within which to write the file.
            label: Name of the file to write.
            n_initial_conditions: Number of initial conditions / timeseries
                to write to the file.
            save_names: Names of variables to save in the output file.
                If None, all provided variables will be saved.
            variable_metadata: Metadata for each variable to be written to the file.
            coords: Coordinate data to be written to the file.
            dataset_metadata: Metadata for the dataset.
        """
        filename = str(Path(path) / label)
        self._save_names = save_names
        self.variable_metadata = variable_metadata
        self.coords = coords
        self.dataset = Dataset(filename, "w", format="NETCDF4")
        self.dataset.createDimension(LEAD_TIME_DIM, None)  # unlimited dimension
        self.dataset.createVariable(LEAD_TIME_DIM, "i8", (LEAD_TIME_DIM,))
        self.dataset.variables[LEAD_TIME_DIM].units = LEAD_TIME_UNITS
        self.dataset.createDimension(IC_DIM, n_initial_conditions)
        self.dataset.createVariable(INIT_TIME, "i8", (IC_DIM,))
        self.dataset.variables[INIT_TIME].units = INIT_TIME_UNITS
        self.dataset.createVariable(VALID_TIME, "i8", (IC_DIM, LEAD_TIME_DIM))
        self.dataset.variables[VALID_TIME].units = INIT_TIME_UNITS
        self._dataset_dims_created = False
        dataset_metadata = copy.copy(dataset_metadata)
        dataset_metadata.title = (
            f"ACE {label.removesuffix('.nc').replace('_', ' ')} data file"
        )
        for key, value in dataset_metadata.as_flat_str_dict().items():
            self.dataset.setncattr(key, value)

    def _get_variable_names_to_save(
        self, *data_varnames: Iterable[str]
    ) -> Iterable[str]:
        return get_all_names(*data_varnames, allowlist=self._save_names)

    def append_batch(
        self,
        data: dict[str, torch.Tensor],
        start_timestep: int,
        batch_time: xr.DataArray,
    ):
        """
        Append a batch of data to the file.

        Args:
            data: Data to be written to file.
            start_timestep: Timestep (lead time dim) at which to start writing.
            batch_time: Time coordinate for each sample in the batch.
        """
        if self.dataset is None:
            return
        n_samples_data = list(data.values())[0].shape[0]
        n_samples_time = batch_time.sizes["sample"]
        if n_samples_data != n_samples_time:
            raise ValueError(
                f"Batch size mismatch, data has {n_samples_data} samples "
                f"and times has {n_samples_time} samples."
            )
        n_times_data = list(data.values())[0].shape[1]
        n_times_time = batch_time.sizes["time"]
        if n_times_data != n_times_time:
            raise ValueError(
                f"Batch time dimension mismatch, data has {n_times_data} times "
                f"and time has {n_times_time} times."
            )

        if not self._dataset_dims_created:
            _dim_info = DIM_INFO_HEALPIX if "face" in self.coords else DIM_INFO_LATLON
            _ordered_names = []
            for dim in _dim_info:
                dim_size = data[next(iter(data.keys()))].shape[dim.index]
                self.dataset.createDimension(dim.name, dim_size)
                if dim.name in self.coords:
                    self.dataset.createVariable(dim.name, "f4", (dim.name,))
                    self.dataset.variables[dim.name][:] = self.coords[dim.name]
                _ordered_names.append(dim.name)
            dims = (IC_DIM, LEAD_TIME_DIM, *_ordered_names)
            self._dataset_dims_created = True

        save_names = self._get_variable_names_to_save(data.keys())
        for variable_name in save_names:
            # define the variable if it doesn't exist
            if variable_name not in self.dataset.variables:
                self.dataset.createVariable(
                    variable_name,
                    "f4",
                    dims,
                    fill_value=np.nan,
                )
                if variable_name in self.variable_metadata:
                    self.dataset.variables[
                        variable_name
                    ].units = self.variable_metadata[variable_name].units
                    self.dataset.variables[
                        variable_name
                    ].long_name = self.variable_metadata[variable_name].long_name
                self.dataset.variables[variable_name].coordinates = " ".join(
                    [INIT_TIME, VALID_TIME]
                )

            data_numpy = data[variable_name].cpu().numpy()
            # Append the data to the variables
            self.dataset.variables[variable_name][
                :,
                start_timestep : start_timestep + data_numpy.shape[1],
                :,
            ] = data_numpy

        # handle time dimensions
        if not hasattr(self.dataset.variables[INIT_TIME], "calendar"):
            self.dataset.variables[INIT_TIME].calendar = batch_time.dt.calendar
        if not hasattr(self.dataset.variables[VALID_TIME], "calendar"):
            self.dataset.variables[VALID_TIME].calendar = batch_time.dt.calendar

        if start_timestep == 0:
            init_times: np.ndarray = batch_time.isel(time=0).values
            init_times_numeric: np.ndarray = cftime.date2num(
                init_times,
                units=self.dataset.variables[INIT_TIME].units,
                calendar=self.dataset.variables[INIT_TIME].calendar,
            )
            self.dataset.variables[INIT_TIME][:] = init_times_numeric
        else:
            init_times_numeric = self.dataset.variables[INIT_TIME][:]
            init_times_numeric = (
                init_times_numeric.filled()
            )  # convert masked array to ndarray
            init_times = cftime.num2date(
                init_times_numeric,
                units=self.dataset.variables[INIT_TIME].units,
                calendar=self.dataset.variables[INIT_TIME].calendar,
            )
        lead_time_microseconds = get_batch_lead_time_microseconds(
            init_times,
            batch_time.values,
        )
        self.dataset.variables[LEAD_TIME_DIM][
            start_timestep : start_timestep + lead_time_microseconds.shape[0]
        ] = lead_time_microseconds

        valid_times_numeric: np.ndarray = cftime.date2num(
            batch_time.values,
            units=self.dataset.variables[VALID_TIME].units,
            calendar=self.dataset.variables[VALID_TIME].calendar,
        )
        self.dataset.variables[VALID_TIME][
            :,
            start_timestep : start_timestep + lead_time_microseconds.shape[0],
        ] = valid_times_numeric

        self.dataset.sync()  # Flush the data to disk

    def flush(self):
        """
        Flush the data to disk.
        """
        self.dataset.sync()


def get_batch_lead_time_microseconds(
    init_time: npt.NDArray[cftime.datetime], batch_time: npt.NDArray[cftime.datetime]
) -> npt.NDArray[np.int64]:
    """
    Get the lead time in seconds for the batch.
    Assert that they are the same for each sample.

    Args:
        init_time: Initialization time for each sample in the batch.
        batch_time: Array of time coordinates for each sample in the batch.

    Returns:
        Lead time in microseconds for the batch
    """
    if init_time.shape[0] != batch_time.shape[0]:
        raise ValueError(
            f"Number of init times ({len(init_time)}) must "
            f"match number of batch times ({len(batch_time)})"
        )
    # Carry out timedelta arithmetic in NumPy arrays to avoid xarray's automatic
    # casting of datetime.timedelta objects to timedelta64[ns] values, which would
    # unnecessarily limit the lead time coordinate to containing values less than
    # ~292 years. See
    # https://numpy.org/doc/stable/reference/arrays.datetime.html#datetime-units
    # for more details on the limits of various precision timedeltas.
    lead_time: npt.NDArray[datetime.timedelta] = (  # type: ignore
        batch_time - init_time[:, None]
    )
    lead_time_microseconds: npt.NDArray[np.int64] = (
        lead_time // datetime.timedelta(microseconds=1)
    ).astype(np.int64)
    if not np.all(lead_time_microseconds == lead_time_microseconds[0, :]):
        raise ValueError("Lead times are not the same for each sample in the batch.")
    return lead_time_microseconds[0, :]



# This script is used to compute a training dataset from the "raw"
# FV3GFS data stored in zarr form on GCS.

# The resulting dataset is about 194GB (the input is about 2.5TB). Running this script
# on my 8-CPU VM takes about 2.5 hours. See "compute_dataset_fv3gfs_argo_workflow.yaml"
# for a workflow which parallelizes this script across the 11-member ensemble and runs
# it on our GKE cluster.

import abc
import dataclasses
import logging
import os
import sys
import time
from typing import Dict, List, Mapping, MutableMapping, Optional, Sequence, Tuple, Union

import click
import dacite
import fsspec
import numpy as np
import xarray as xr
import yaml

try:
    from xtorch_harmonics import roundtrip_filter
except ModuleNotFoundError:

    def roundtrip_filter(*args, **kwargs):
        raise ModuleNotFoundError("xtorch_harmonics")


sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from get_stats import StatsConfig

# constants are defined as in FV3GFS model
# https://github.com/ai2cm/fv3gfs-fortran/blob/master/FMS/constants/constants.F90
LATENT_HEAT_OF_VAPORIZATION = 2.5e6  # J/kg
GRAVITY = 9.80665  # m/s^2

SPECIFIC_TOTAL_WATER = "specific_total_water"
# following name is for the water path computed by vertically integrating the
# lat-lon 3D variables. This is the water path we want to use for training. It is not
# necessarily exactly equal to the precipitable water path computed online by FV3GFS.
TOTAL_WATER_PATH = "total_water_path"
SURFACE_FROZEN_PRECIPITATION_NAME = "total_frozen_precipitation_rate"


@dataclasses.dataclass
class StandardDimMapping:
    longitude_dim: str = "grid_xt"
    latitude_dim: str = "grid_yt"
    time_dim: str = "time"


@dataclasses.dataclass
class StandardNameMapping(StandardDimMapping):
    vertical_dim: str = "pfull"
    vertical_interface_dim: str = "phalf"
    surface_pressure: str = "PRESsfc"
    latent_heat_flux: str = "LHTFLsfc"
    precip_rate: str = "PRATEsfc"
    surface_snow_rate: str = "SNOWsfc"
    surface_ice_rate: str = "ICEsfc"
    surface_graupel_rate: str = "GRAUPELsfc"
    total_frozen_precip_rate: str = "total_frozen_precipitation_rate"
    precipitable_water_path: str = "precipitable_water_path"
    pressure_thickness: str = "pressure_thickness_of_atmospheric_layer"
    air_temperature: str = "air_temperature"
    specific_humidity: str = "specific_humidity"
    cloud_water_mixing_ratio: str = "cloud_water_mixing_ratio"
    cloud_ice_mixing_ratio: str = "cloud_ice_mixing_ratio"
    graupel_mixing_ratio: str = "graupel_mixing_ratio"
    rain_mixing_ratio: str = "rain_mixing_ratio"
    snow_mixing_ratio: str = "snow_mixing_ratio"
    northward_wind: str = "northward_wind"
    eastward_wind: str = "eastward_wind"
    surface_evaporation_rate: str = "surface_evaporation_rate"
    land_fraction: str = "land_fraction"
    ocean_fraction: str = "ocean_fraction"
    sea_ice_fraction: str = "sea_ice_fraction"
    vertical_dim_land: str = "zfull_soil"
    height_thickness: str = "height_thickness_of_land_layer"
    total_moisture_content_of_soil_layer: str = "total_moisture_content_of_soil_layer"
    hybrid_level_coeffs: List[str] = dataclasses.field(default_factory=list)
    additional_vertically_resolved_names: List[str] = dataclasses.field(
        default_factory=list
    )
    land_names_to_vertically_coarsen_by_height_weighting: List[str] = dataclasses.field(
        default_factory=list
    )
    land_names_to_vertically_coarsen_by_sum: List[str] = dataclasses.field(
        default_factory=list
    )

    def __post_init__(self):
        self.horizontal_dims: List[str] = [self.longitude_dim, self.latitude_dim]

        self.specific_total_water = SPECIFIC_TOTAL_WATER
        self.total_water_path = TOTAL_WATER_PATH
        self.total_frozen_precip_rate_output_name = SURFACE_FROZEN_PRECIPITATION_NAME
        self.pwat_tendency = f"tendency_of_{self.total_water_path}"
        self.time_derivative_names = [self.total_water_path]

        self.vertically_resolved: List[str] = [
            self.specific_total_water,
            self.air_temperature,
            self.northward_wind,
            self.eastward_wind,
        ] + self.additional_vertically_resolved_names

        self.vertically_resolved_names_land: List[str] = (
            self.land_names_to_vertically_coarsen_by_height_weighting
            + self.land_names_to_vertically_coarsen_by_sum
        )

        # variables to drop after all derived variables are computed
        self.dropped_variables: List[str] = (
            self.water_species
            + self.vertically_resolved
            + [self.pressure_thickness, self.vertical_dim]
            + self.vertically_resolved_names_land
        )
        for name in [
            self.precipitable_water_path,
            self.surface_graupel_rate,
            self.surface_ice_rate,
            self.surface_snow_rate,
        ]:
            if name.lower() != "none":
                self.dropped_variables.append(name)
        if self.vertically_resolved_names_land:
            self.dropped_variables.append(self.vertical_dim_land)

    @property
    def water_species(self) -> List[str]:
        return [
            item
            for item in [
                self.specific_humidity,
                self.cloud_water_mixing_ratio,
                self.cloud_ice_mixing_ratio,
                self.graupel_mixing_ratio,
                self.rain_mixing_ratio,
                self.snow_mixing_ratio,
            ]
            if item.lower() != "none"
        ]

    @property
    def frozen_precipitation_species(self) -> List[str]:
        if self.total_frozen_precip_rate.lower() != "none":
            # if total frozen precip rate is available, just use that
            return [self.total_frozen_precip_rate]
        else:
            # return all frozen precip species
            return [
                item
                for item in [
                    self.surface_graupel_rate,
                    self.surface_ice_rate,
                    self.surface_snow_rate,
                ]
                if item.lower() != "none"
            ]


@dataclasses.dataclass
class DLWPNameMapping(StandardNameMapping):
    longitude_dim: str = "longitude"
    latitude_dim: str = "latitude"
    face_dim: str = "face"
    width_dim: str = "width"
    height_dim: str = "height"

    def __post_init__(self):
        super().__post_init__()

        self.horizontal_dims: List[str] = [
            self.face_dim,
            self.width_dim,
            self.height_dim,
        ]
        self.lat_lon_dims: List[str] = [self.latitude_dim, self.longitude_dim]


@dataclasses.dataclass
class _ChunkingConfig(abc.ABC):
    time_dim: int = 1

    @abc.abstractmethod
    def get_chunks(self, standard_names: StandardDimMapping) -> Dict[str, int]: ...


@dataclasses.dataclass
class ChunkingConfig(_ChunkingConfig):
    latitude_dim: int = -1
    longitude_dim: int = -1

    def get_chunks(self, standard_names: StandardDimMapping) -> Dict[str, int]:
        return {
            standard_names.time_dim: self.time_dim,
            standard_names.longitude_dim: self.longitude_dim,
            standard_names.latitude_dim: self.latitude_dim,
        }


@dataclasses.dataclass
class DLWPChunkingConfig(_ChunkingConfig):
    face_dim: int = -1
    width_dim: int = -1
    height_dim: int = -1

    def get_chunks(self, standard_names: StandardDimMapping) -> Dict[str, int]:
        dlwp_names = standard_names
        if not isinstance(dlwp_names, DLWPNameMapping):
            raise TypeError(
                "Expected DLWPChunkingConfig to be passed type of DLWPNameMapping."
            )
        chunks = {
            dlwp_names.time_dim: self.time_dim,
            dlwp_names.face_dim: self.face_dim,
            dlwp_names.width_dim: self.width_dim,
            dlwp_names.height_dim: self.height_dim,
        }
        return chunks


@dataclasses.dataclass
class DatasetComputationConfig:
    """Configuration of computation details for an FME reference dataset.

    Parameters:
        reference_vertical_coordinate_file: path to netCDF file containing
            vertical coordinate definition for the reference simulation.
        vertical_coarsening_indices: list of tuples defining the ranges of
            reference levels that go into each vertically coarsened layer.
        variable_sources: mapping of zarr store names, e.g. "full_state.zarr",
            to lists of variables to extract from each.
        validate_vertical_coarsening_indices: (optional) whether to check that
            vertical coarsening indices span the full atmosphere and do not
            overlap (default True).
        n_split: number of steps to split the computation over across time.
        roundtrip_fraction_kept: (optional) fraction of spherical harmonics to
            keep in roundtrip transform. Must be between 0 and 1. If omitted,
            the default, no roundtrip transform is applied.
        renaming: (optional) mapping of names in dataset to renamed output
        standard_names: (optional) mapping of standard names to corresponding
            names of variables in the dataset.
        chunking: (optional) mapping of standard dimension names to desired
            output inner chunk sizes. Defaults to a chunk size of 1 along
            the time dimension.
        sharding: (optional) mapping of standard dimension names to desired
            output shard sizes. Defaults to a shard size of 360 along the time
            dimension. If None, then an unsharded zarr store will be written
            with chunks as specified in ``chunking``.
        time_invariant_dir: (optional) path to directory containing time-invariant data
            This option is used for E3SMv2 dataset.
        vertical_coarsening_indices_land: (optional) list of tuples defining the ranges
            of reference levels that go into each vertically coarsened layer for the
            land model variables.
        validate_vertical_coarsening_indices_land: (optional) whether to check that
            land vertical coarsening indices span the full depth of the land model
            and do not overlap (default True).
        reference_vertical_coordinate_file_land: (optional) path to netCDF file
            containing vertical coordinate definition for the land model of the
            reference simulation.
        mask_soil_moisture: (optional) whether to mask soil moisture content using soil
            temperature. This is useful for CM4 dataset, where soil moisture content is
            zero instead of NaN over the oceans.
    """

    reference_vertical_coordinate_file: str
    vertical_coarsening_indices: Sequence[Tuple[int, int]]
    variable_sources: Mapping[str, Sequence[str]]
    validate_vertical_coarsening_indices: bool = True
    n_split: int = 65
    renaming: Mapping[str, str] = dataclasses.field(default_factory=dict)
    roundtrip_fraction_kept: Optional[float] = None
    standard_names: Union[StandardNameMapping, DLWPNameMapping] = dataclasses.field(
        default_factory=StandardNameMapping
    )
    chunking: Union[ChunkingConfig, DLWPChunkingConfig] = dataclasses.field(
        default_factory=lambda: ChunkingConfig(time_dim=1)
    )
    sharding: Optional[Union[ChunkingConfig, DLWPChunkingConfig]] = dataclasses.field(
        default_factory=lambda: ChunkingConfig(time_dim=360)
    )
    time_invariant_dir: Optional[str] = None
    vertical_coarsening_indices_land: Optional[Sequence[Tuple[int, int]]] = None
    validate_vertical_coarsening_indices_land: bool = True
    reference_vertical_coordinate_file_land: Optional[str] = None
    mask_soil_moisture: bool = False


@dataclasses.dataclass
class DatasetConfig:
    """Dataset provenance for a set of reference simulations.

    Parameters:
        runs: mapping of short names to full paths of reference datasets.
        output_directory: path to place output of computation script.
        dataset_computation: configuration details for dataset
            computation.
        stats_config: configuration to retrieve statistics dataset
    """

    runs: Mapping[str, str]
    data_output_directory: str
    dataset_computation: DatasetComputationConfig
    stats: StatsConfig

    @classmethod
    def from_file(cls, path: str) -> "DatasetConfig":
        with open(path, "r") as file:
            data = yaml.safe_load(file)

        return dacite.from_dict(
            data_class=cls, data=data, config=dacite.Config(cast=[tuple], strict=True)
        )


def weighted_mean(da: xr.DataArray, weights: xr.DataArray, dims) -> xr.DataArray:
    """Compute weighted mean of xr.DataArray."""
    return (da * weights).sum(dims, skipna=False) / weights.sum(dims)


def get_dataset_urls(
    config: DatasetComputationConfig, run_directory: str
) -> MutableMapping[str, str]:
    return {k: os.path.join(run_directory, k) for k in config.variable_sources}


def open_datasets(
    config: DatasetComputationConfig, urls: MutableMapping[str, str]
) -> xr.Dataset:
    datasets = []
    for store, names in config.variable_sources.items():
        url = urls[store]
        ds = xr.open_zarr(url, decode_timedelta=True)[names]
        datasets.append(ds)
    return xr.merge(datasets, compat="equals")


def get_coarse_ak_bk(
    url: str,
    interface_indices: Sequence[Tuple[int, int]],
    z_dim="xaxis_1",
    time_dim="Time",
    dtype=np.float32,
) -> xr.Dataset:
    """Return dataset with scalar ak and bk coordinates that define coarse interfaces.

    Args:
        url: path to netCDF file with ak and bk variables in format output by FV3GFS.
        interface_indices: list of tuples of indices of the interfaces in the vertical.
        z_dim: name of dimension along which ak and bk are defined.
        time_dim: name of time dimension.
        dtype: data type (e.g., np.float32) for ak and bk

    Returns:
        xr.Dataset with ak and bk variables as scalars labeled as ak_0, bk_0, etc.

    Note:
        The ak and bk variables will have one more vertical level than the other 3D
        variables since they represent the interfaces between levels.
    """
    with fsspec.open(url) as f:
        vertical_coordinate = xr.open_dataset(f, decode_timedelta=False).load()

    vertical_coordinate = vertical_coordinate.astype(dtype)

    # squeeze out the singleton time dimension
    vertical_coordinate = vertical_coordinate.squeeze()
    data = {}
    for i, (start, end) in enumerate(interface_indices):
        data[f"ak_{i}"] = vertical_coordinate["ak"].isel({z_dim: start})
        data[f"bk_{i}"] = vertical_coordinate["bk"].isel({z_dim: start})
        if i == len(interface_indices) - 1:
            data[f"ak_{i + 1}"] = vertical_coordinate["ak"].isel({z_dim: end})
            data[f"bk_{i + 1}"] = vertical_coordinate["bk"].isel({z_dim: end})
    for i in range(len(interface_indices) + 1):
        data[f"ak_{i}"].attrs["units"] = "Pa"
        data[f"bk_{i}"].attrs["units"] = ""  # unitless quantity
        for name in ["ak", "bk"]:
            data[f"{name}_{i}"] = data[f"{name}_{i}"].drop_vars([z_dim, time_dim])
    return xr.Dataset(data)


def compute_ocean_fraction(
    ds: xr.Dataset,
    output_name: str,
    land_fraction_name: str,
    sea_ice_fraction_name: str,
) -> xr.Dataset:
    """Compute latent heat flux, if needed."""
    if output_name in ds.data_vars:
        # if ocean_fraction is already computed, assume that NaNs have been handled
        return ds
    ds[sea_ice_fraction_name] = ds[sea_ice_fraction_name].fillna(0.0)
    ocean_fraction = 1 - ds[sea_ice_fraction_name] - ds[land_fraction_name]
    negative_ocean = xr.where(ocean_fraction < 0, ocean_fraction, 0)
    ocean_fraction -= negative_ocean
    ds["sea_ice_fraction"] += negative_ocean
    ocean_fraction.attrs["units"] = "unitless"
    ocean_fraction.attrs["long_name"] = "fraction of grid cell area occupied by ocean"
    return ds.assign({output_name: ocean_fraction})


def compute_latent_heat_flux(
    ds: xr.Dataset,
    output_name: str,
    evaporation_name: Optional[str] = None,
) -> xr.Dataset:
    """Compute latent heat flux, if needed."""
    if output_name in ds.data_vars:
        return ds
    assert (
        evaporation_name is not None
    ), f"{output_name} not found in ds, evaporation_name must be provided."
    latent_heat_flux = ds[evaporation_name] * LATENT_HEAT_OF_VAPORIZATION
    latent_heat_flux.attrs["units"] = "W/m^2"
    latent_heat_flux.attrs["long_name"] = "Latent heat flux"
    return ds.assign({output_name: latent_heat_flux}).drop_vars(evaporation_name)


def compute_specific_total_water(
    ds: xr.Dataset, water_condensate_names: Sequence[str], output_name: str
) -> xr.Dataset:
    """Compute specific total water from individual water species."""
    specific_total_water: xr.DataArray = sum(
        [ds[name] for name in water_condensate_names]
    )
    specific_total_water.attrs["units"] = "kg/kg"
    specific_total_water.attrs["long_name"] = output_name.replace("_", " ")
    return ds.assign({output_name: specific_total_water})


def compute_frozen_precipitation_rate(
    ds: xr.Dataset, frozen_precip_names: Sequence[str], output_name: str
) -> xr.Dataset:
    """Compute the total surface frozen precipitation rate."""
    frozen_precip: xr.DataArray = sum([ds[name] for name in frozen_precip_names])
    frozen_precip.attrs["units"] = ds[frozen_precip_names[0]].units
    frozen_precip.attrs["long_name"] = "Total surface frozen precipitation rate"
    return ds.assign({output_name: frozen_precip})


def compute_pressure_thickness(
    ds: xr.Dataset,
    vertical_coordinate_file: str,
    vertical_dim_name: str,
    surface_pressure_name: str,
    output_name: str,
    z_dim: str = "xaxis_1",
):
    if output_name in ds.data_vars:
        return ds

    with fsspec.open(vertical_coordinate_file) as f:
        vertical_coordinate = xr.open_dataset(f, decode_timedelta=False).load()

    vertical_coordinate = vertical_coordinate.astype(ds[surface_pressure_name].dtype)

    # squeeze out the singleton time dimension
    vertical_coord = vertical_coordinate.squeeze(drop=True)

    sfc_pressure = ds[surface_pressure_name].expand_dims(
        {z_dim: vertical_coord[z_dim]}, axis=3
    )
    phalf = sfc_pressure * vertical_coord["bk"] + vertical_coord["ak"]

    thickness = (
        phalf.diff(dim=z_dim)
        .rename({z_dim: vertical_dim_name})
        .rename(output_name)
        .assign_coords(
            {vertical_dim_name: (vertical_dim_name, ds[vertical_dim_name].values)}
        )
    )

    thickness.attrs["units"] = "Pa"
    thickness.attrs["long_name"] = output_name.replace("_", " ")
    return ds.assign({output_name: thickness})


def validate_vertical_coarsening_indices(
    vertical_dim_size: int,
    interface_indices: Sequence[Tuple[int, int]],
    component: str,
    control_flag: str,
) -> None:
    """Check that vertical coarsening indices span the full vertical dimension of
    the model component and do not overlap.

    Args:
        vertical_dim_size: size of the vertical dimension in the input dataset.
        interface_indices: provided vertical coarsening indices.
        component: model component the vertical dimension corresponds to (only
            relevant for providing a clearer error message).
        control_flag: relevant control flag on DatasetComputationConfig to disable
            this check if desired (only relevant for providing a clearer error
            message).
    """
    expected_covered_indices = list(range(vertical_dim_size))
    covered_indices = []
    for start, end in interface_indices:
        covered_indices.extend(list(range(start, end)))

    if covered_indices != expected_covered_indices:
        raise ValueError(
            f"Provided {component} vertical coarsening indices {interface_indices!r} "
            f"do not exactly span all {vertical_dim_size} vertical levels of the "
            f"input dataset and/or consist of overlapping ranges of levels. If your "
            f"intent is to use incomplete or overlapping levels, you may disable this "
            f"check by setting {control_flag} in DatasetComputationConfig to False. "
            f"If this is not your intent, double check that you are using the proper "
            f"coarsening indices and, if relevant, the vertical coordinate reference "
            f"file for this reference model component configuration."
        )


def compute_vertical_coarsening(
    ds: xr.Dataset,
    vertically_resolved_names: Sequence[str],
    interface_indices: Sequence[Tuple[int, int]],
    dim: str,
    pressure_thickness_name: str,
    validate_indices: bool,
) -> xr.Dataset:
    """Compute vertical coarsening of 3D variables by mass-weighted mean. Outputs are
    saved as new variables in the dataset with the name '{name}_{i}' where i is the
    new coarse vertical level index."""

    if validate_indices:
        validate_vertical_coarsening_indices(
            ds.sizes[dim],
            interface_indices,
            "atmosphere",
            "validate_vertical_coarsening_indices",
        )

    coarsened_arrays = {}
    for i, (start, end) in enumerate(interface_indices):
        pressure_thickness = ds[pressure_thickness_name].isel({dim: slice(start, end)})
        for name in vertically_resolved_names:
            array_slice = ds[name].isel({dim: slice(start, end)})
            coarsened_da = weighted_mean(array_slice, pressure_thickness, dim)
            current_long_name = array_slice.long_name
            coarsened_da.attrs["long_name"] = current_long_name + f" level-{i}"
            coarsened_arrays[f"{name}_{i}"] = coarsened_da
    return ds.assign(coarsened_arrays)


def compute_vertical_coarsening_land(
    ds: xr.Dataset,
    vertically_resolved_names: Sequence[str],
    interface_indices: Optional[Sequence[Tuple[int, int]]],
    vertical_coordinate_file: Optional[str],
    dim: str,
    height_thickness_name: str,
    summed_variables: Sequence[str],
    validate_indices: bool,
    mask_soil_moisture: bool = False,
) -> xr.Dataset:
    """Compute vertical coarsening of 3D land variables by height-weighted mean or
    unweighted sum. Outputs are saved as new variables in the dataset with the
    name '{name}_{i}' where i is the new coarse vertical level index. Variables are
    coarsened by a height-weighted mean by default. Variables listed in
    summed_variables are coarsened using an unweighted sum. This is useful, for
    example, for the total_moisture_content_of_soil_layer in LM4, which has units
    of total moisture (kg / m^2).
    """

    assert interface_indices is not None, (
        "Land variables for coarsening are provided, but there"
        "are not corresponding coarsening indices"
    )

    assert vertical_coordinate_file is not None, (
        "Land variables for coarsening are provided, but there"
        "is not a corresponding reference file"
    )

    if not vertically_resolved_names:
        return ds

    if validate_indices:
        validate_vertical_coarsening_indices(
            ds.sizes[dim],
            interface_indices,
            "land",
            "validate_vertical_coarsening_indices_land",
        )

    with fsspec.open(vertical_coordinate_file) as f:
        thickness = xr.open_dataset(f, decode_timedelta=False).load()

    coarsened_arrays = {}

    if mask_soil_moisture and (
        "total_moisture_content_of_soil_layer" in vertically_resolved_names
        and "temperature_of_soil_layer" in vertically_resolved_names
    ):
        # Mask out soil moisture content when temperature is NaN
        logging.info("Masking soil moisture content using temperature")
        ds["total_moisture_content_of_soil_layer"] = ds[
            "total_moisture_content_of_soil_layer"
        ].where(ds.temperature_of_soil_layer.notnull())

    for i, (start, end) in enumerate(interface_indices):
        height_thickness = thickness[height_thickness_name].isel(
            {dim: slice(start, end)}
        )
        for name in vertically_resolved_names:
            array_slice = ds[name].isel({dim: slice(start, end)})

            # some land variables are total quantity in layer so just need to be summed
            if name in summed_variables:
                coarsened_da = array_slice.sum(dim, skipna=False)
            else:
                coarsened_da = weighted_mean(
                    array_slice, height_thickness.astype(array_slice.dtype), dim
                )

            current_long_name = array_slice.long_name
            coarsened_da.attrs["long_name"] = current_long_name + f" level-{i}"
            coarsened_arrays[f"{name}_{i}"] = coarsened_da
    return ds.assign(coarsened_arrays)


def compute_tendencies(
    ds: xr.Dataset, time_derivative_names: Sequence[str], dim: str
) -> xr.Dataset:
    """Compute backward difference over time dimension. The tendency variables
    will be NaNs for the first timestep in the output dataset."""
    # this code does not assume that all time steps are equally spaced
    timestep_seconds = (ds[dim].diff(dim) / np.timedelta64(1, "s")).astype("float32")
    tendencies = {}
    for name in time_derivative_names:
        tendency = ds[name].diff(dim) / timestep_seconds
        tendency.attrs["units"] = f"{ds[name].units}/s"
        tendency.attrs["long_name"] = f"time derivative of {ds[name].long_name}"
        tendencies[f"tendency_of_{name}"] = tendency
    return ds.assign(tendencies)


def compute_column_advective_moisture_tendency(
    ds: xr.Dataset,
    pwat_tendency: str,
    latent_heat_flux: str,
    precip: str,
    latent_heat_of_vaporization: float,
) -> xr.Dataset:
    evaporation = ds[latent_heat_flux] / latent_heat_of_vaporization
    advective_tendency = ds[pwat_tendency] - evaporation + ds[precip]
    long_name = "tendency of total water path due to advection"
    advective_tendency.attrs["long_name"] = long_name
    return ds.assign({f"{pwat_tendency}_due_to_advection": advective_tendency})


def compute_column_moisture_integral(
    ds: xr.Dataset,
    input_name: str,
    output_name: str,
    pressure_thickness_name: str,
    dim: str,
) -> xr.Dataset:
    """Compute the column integral of a mass mixing ratio."""
    column_integral = (ds[input_name] * ds[pressure_thickness_name]).sum(dim) / GRAVITY
    column_integral.attrs["units"] = "kg/m^2"
    column_integral.attrs["long_name"] = output_name.replace("_", " ")
    return ds.assign({output_name: column_integral})


def assert_column_integral_of_moisture_is_conserved(
    ds: xr.Dataset, precipitable_water_path_name: str, total_water_path_name: str
) -> None:
    """Assert that the column integral of 'specific_total_water' is close to the
    precipitable water path that is computed online on native grid by FV3GFS. Note
    that there are pretty large difference (rtol=1e-1) which is likely due to the
    fregrid tool doing area-weighted average instead of mass-weighted."""
    expected_pwat = ds[precipitable_water_path_name]
    integrated_pwat = ds[total_water_path_name]
    logging.info(
        f"Mean absolute difference between expected and integrated pwat [kg/m^2]: "
        f"{np.abs(expected_pwat - integrated_pwat).mean().values}"
    )
    xr.testing.assert_allclose(integrated_pwat, expected_pwat, rtol=1e-1, atol=1e-3)


def assert_global_dry_air_mass_conservation(
    ds: xr.Dataset,
    dims: List[str],
    surface_pressure_name: str,
    total_water_path_name: str,
    latitude_dim: str,
    time_dim: str,
) -> None:
    """Assert that the tendency of global average surface pressure (due to dry air
    only) is close to zero. I.e. dry air mass is conserved."""
    column_dry_air_mass = (
        ds[surface_pressure_name] - ds[total_water_path_name] * GRAVITY
    )
    if latitude_dim in dims:
        weights = np.cos(np.deg2rad(ds[latitude_dim]))
        global_dry_air_mass = column_dry_air_mass.weighted(weights).mean(dim=dims)
    else:
        global_dry_air_mass = column_dry_air_mass.mean(dim=dims)

    global_dry_air_mass_tendency = global_dry_air_mass.diff(time_dim)
    logging.info(
        f"Mean absolute global dry air pressure tendency [Pa]: "
        f"{np.abs(global_dry_air_mass_tendency).mean().values}"
    )
    xr.testing.assert_allclose(
        global_dry_air_mass_tendency,
        xr.zeros_like(global_dry_air_mass_tendency),
        atol=1e-3,
    )


def assert_global_moisture_conservation(
    ds: xr.Dataset,
    dims: List[str],
    latitude_dim: str,
    total_water_path_name: str,
    latent_heat_flux_name: str,
    latent_heat_of_vaporization: float,
    precip_rate_name: str,
    time_dim: str,
) -> None:
    """Assert that the tendency of global average column integrated moisture is equal
    to the global average flux of moisture through the surface."""
    integrated_pwat = ds[total_water_path_name]
    weights = np.cos(np.deg2rad(ds[latitude_dim]))
    global_moisture = integrated_pwat.weighted(weights).mean(dim=dims)
    timestep_seconds = ds[time_dim].diff(time_dim) / np.timedelta64(1, "s")
    actual_global_moisture_tendency = global_moisture.diff(time_dim) / timestep_seconds
    evap_minus_precip = (
        ds[latent_heat_flux_name] / latent_heat_of_vaporization - ds[precip_rate_name]
    )
    expected_global_moisture_tendency = (
        evap_minus_precip.weighted(weights).mean(dim=dims).isel(time=slice(1, None))
    )
    diff = actual_global_moisture_tendency - expected_global_moisture_tendency
    logging.info(
        f"Mean absolute global moisture non-conservative source [kg/m^2/s]: "
        f"{np.abs(diff).mean().values}"
    )
    xr.testing.assert_allclose(
        expected_global_moisture_tendency, actual_global_moisture_tendency
    )


def construct_lazy_dataset(
    config: DatasetComputationConfig, run_directory: str
) -> xr.Dataset:
    standard_names = config.standard_names
    urls = get_dataset_urls(config, run_directory)
    ds = open_datasets(config, urls)
    for var in ds:
        del ds[var].encoding["chunks"]
        del ds[var].encoding["preferred_chunks"]
    logging.info(f"Input dataset size is {ds.nbytes / 1e9} GB")

    if config.roundtrip_fraction_kept is not None:
        ds = roundtrip_filter(
            ds,
            lat_dim=standard_names.latitude_dim,
            lon_dim=standard_names.longitude_dim,
            fraction_modes_kept=config.roundtrip_fraction_kept,
        )
    ds = compute_ocean_fraction(
        ds,
        output_name=standard_names.ocean_fraction,
        land_fraction_name=standard_names.land_fraction,
        sea_ice_fraction_name=standard_names.sea_ice_fraction,
    )
    ds = compute_latent_heat_flux(
        ds,
        output_name=standard_names.latent_heat_flux,
        evaporation_name=standard_names.surface_evaporation_rate,
    )
    ds = compute_specific_total_water(
        ds,
        water_condensate_names=standard_names.water_species,
        output_name=standard_names.specific_total_water,
    )
    ds = compute_frozen_precipitation_rate(
        ds,
        frozen_precip_names=standard_names.frozen_precipitation_species,
        output_name=standard_names.total_frozen_precip_rate_output_name,
    )
    ds = compute_pressure_thickness(
        ds,
        vertical_coordinate_file=config.reference_vertical_coordinate_file,
        vertical_dim_name=standard_names.vertical_dim,
        surface_pressure_name=standard_names.surface_pressure,
        output_name=standard_names.pressure_thickness,
    )
    ds = compute_vertical_coarsening(
        ds,
        vertically_resolved_names=standard_names.vertically_resolved,
        interface_indices=config.vertical_coarsening_indices,
        dim=standard_names.vertical_dim,
        pressure_thickness_name=standard_names.pressure_thickness,
        validate_indices=config.validate_vertical_coarsening_indices,
    )

    if standard_names.vertically_resolved_names_land:
        ds = compute_vertical_coarsening_land(
            ds,
            vertically_resolved_names=standard_names.vertically_resolved_names_land,
            interface_indices=config.vertical_coarsening_indices_land,
            vertical_coordinate_file=config.reference_vertical_coordinate_file_land,
            dim=standard_names.vertical_dim_land,
            height_thickness_name=standard_names.height_thickness,
            summed_variables=standard_names.land_names_to_vertically_coarsen_by_sum,
            validate_indices=config.validate_vertical_coarsening_indices_land,
            mask_soil_moisture=config.mask_soil_moisture,
        )

    ds = compute_column_moisture_integral(
        ds,
        input_name=standard_names.specific_total_water,
        output_name=standard_names.total_water_path,
        pressure_thickness_name=standard_names.pressure_thickness,
        dim=standard_names.vertical_dim,
    )
    ds = compute_tendencies(
        ds,
        time_derivative_names=standard_names.time_derivative_names,
        dim=standard_names.time_dim,
    )
    ds = compute_column_advective_moisture_tendency(
        ds,
        pwat_tendency=standard_names.pwat_tendency,
        latent_heat_flux=standard_names.latent_heat_flux,
        precip=standard_names.precip_rate,
        latent_heat_of_vaporization=LATENT_HEAT_OF_VAPORIZATION,
    )
    ak_bk_ds = get_coarse_ak_bk(
        config.reference_vertical_coordinate_file,
        config.vertical_coarsening_indices,
        dtype=ds[standard_names.surface_pressure].dtype,
    )
    ds = xr.merge([ds, ak_bk_ds])

    if config.sharding is None:
        outer_chunks = config.chunking.get_chunks(standard_names)
    else:
        outer_chunks = config.sharding.get_chunks(standard_names)

    ds = ds.chunk(outer_chunks)

    ds.attrs["history"] = (
        "Dataset computed by full-model/scripts/data_process"
        "/compute_dataset_fv3gfs.py"
        f" script, using following input zarrs: {urls.values()}."
    )
    ds.attrs["vertical_coordinate"] = (
        "The pressure at level interfaces can by computed as "
        "p_i = ak_i + bk_i * PRESsfc, where PRESsfc is the surface pressure and the "
        "p_i pressure corresponds to the interface at the top of the i'th finite "
        "volume layer, counting down from the top of atmosphere."
    )
    ds = ds.rename(config.renaming)
    return ds


def clear_compressors_encoding(ds: xr.Dataset) -> xr.Dataset:
    """Clear "compressors" encoding of the Dataset.

    This ensures that we use zarr's default zstd encoding when writing out the
    Dataset. It also helps avoid errors resulting from lingering zarr v2
    compression encoding parameters that are not compatible with zarr v3.

    Args:
        ds: input xr.Dataset to remove "compressors" encoding from.

    Returns:
        xr.Dataset with the "compressors" key in the encoding dictionary of
        each variable removed, if it exists.
    """
    ds = ds.copy(deep=False)
    for variable in {**ds.coords, **ds.data_vars}.values():
        variable.encoding.pop("compressors", None)
    return ds


@click.command()
@click.option("--config", help="Path to dataset configuration YAML file.")
@click.option("--run-directory", help="Path to reference run directory.")
@click.option("--output-store", help="Path to output zarr store.")
@click.option("--debug", is_flag=True, help="Print metadata instead of writing output.")
@click.option("--subsample", is_flag=True, help="Subsample the data before writing.")
@click.option("--check-conservation", is_flag=True, help="Check conservation.")
def main(
    config,
    run_directory,
    output_store,
    debug,
    subsample,
    check_conservation,
):
    # Import distributed and xpartition here to allow testing functions in
    # the fme environment.
    import distributed
    import xpartition  # noqa: F401

    logging.basicConfig(level=logging.INFO)
    distributed.Client(n_workers=16)

    config = DatasetConfig.from_file(config).dataset_computation
    logging.info(f"--run-directory is {run_directory}")
    logging.info(f"--output-store is {output_store}")
    standard_names = config.standard_names
    xr.set_options(keep_attrs=True)
    ds = construct_lazy_dataset(config, run_directory)
    if subsample:
        ds = ds.isel(time=slice(10, 13))
    if check_conservation:
        assert_column_integral_of_moisture_is_conserved(
            ds,
            precipitable_water_path_name=standard_names.precipitable_water_path,
            total_water_path_name=standard_names.total_water_path,
        )
        assert_global_dry_air_mass_conservation(
            ds,
            dims=standard_names.horizontal_dims,
            surface_pressure_name=standard_names.surface_pressure,
            total_water_path_name=standard_names.total_water_path,
            latitude_dim=standard_names.latitude_dim,
            time_dim=standard_names.time_dim,
        )
        assert_global_moisture_conservation(
            ds,
            dims=standard_names.horizontal_dims,
            latitude_dim=standard_names.latitude_dim,
            total_water_path_name=standard_names.total_water_path,
            latent_heat_flux_name=standard_names.latent_heat_flux,
            latent_heat_of_vaporization=LATENT_HEAT_OF_VAPORIZATION,
            precip_rate_name=standard_names.precip_rate,
            time_dim=standard_names.time_dim,
        )
    ds = ds.drop_vars(standard_names.dropped_variables)

    if config.sharding is None:
        inner_chunks = None
    else:
        inner_chunks = config.chunking.get_chunks(standard_names)

    ds = clear_compressors_encoding(ds)

    logging.info(f"Output dataset size is {ds.nbytes / 1e9} GB")
    if debug:
        with xr.set_options(display_max_rows=500):
            logging.info(ds)
    else:
        ds.partition.initialize_store(output_store, inner_chunks=inner_chunks)
        for i in range(config.n_split):
            segment_number = f"{i + 1} / {config.n_split}"
            logging.info(f"Writing segment {segment_number}")
            segment_time = time.time()
            ds.partition.write(
                output_store,
                config.n_split,
                [config.standard_names.time_dim],
                i,
                collect_variable_writes=True,
            )
            segment_time = time.time() - segment_time
            logging.info(f"Segment {segment_number} time: {segment_time:0.2f} seconds")


if __name__ == "__main__":
    main()



# This module is derived from the train.py module in the following repository:
# https://github.com/NVlabs/FourCastNet. The corresponding license is
# provided below.

# BSD 3-Clause License
#
# Copyright (c) 2022, FourCastNet authors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The code was authored by the following people:
#
# Jaideep Pathak - NVIDIA Corporation
# Shashank Subramanian - NERSC, Lawrence Berkeley National Laboratory
# Peter Harrington - NERSC, Lawrence Berkeley National Laboratory
# Sanjeev Raja - NERSC, Lawrence Berkeley National Laboratory
# Ashesh Chattopadhyay - Rice University
# Morteza Mardani - NVIDIA Corporation
# Thorsten Kurth - NVIDIA Corporation
# David Hall - NVIDIA Corporation
# Zongyi Li - California Institute of Technology, NVIDIA Corporation
# Kamyar Azizzadenesheli - Purdue University
# Pedram Hassanzadeh - Rice University
# Karthik Kashinath - NVIDIA Corporation
# Animashree Anandkumar - California Institute of Technology, NVIDIA Corporation

import dataclasses
import logging
import os
from collections.abc import Callable, Sequence

import dacite
import torch
import xarray as xr

import fme
import fme.core.logging_utils as logging_utils
from fme.ace.aggregator import (
    OneStepAggregator,
    OneStepAggregatorConfig,
    TrainAggregator,
)
from fme.ace.aggregator.inference.main import (
    InferenceEvaluatorAggregator,
    InferenceEvaluatorAggregatorConfig,
)
from fme.ace.data_loading.batch_data import BatchData, PairedData, PrognosticState
from fme.ace.stepper import TrainOutput
from fme.ace.train.train_config import TrainBuilders, TrainConfig
from fme.core.cli import prepare_config, prepare_directory
from fme.core.dataset_info import DatasetInfo
from fme.core.derived_variables import get_derived_variable_metadata
from fme.core.dicts import to_flat_dict
from fme.core.distributed import Distributed
from fme.core.generics.data import InferenceDataABC
from fme.core.generics.trainer import (
    AggregatorBuilderABC,
    TrainConfigProtocol,
    Trainer,
    inference_one_epoch,
)
from fme.core.typing_ import TensorDict, TensorMapping


def build_trainer(builder: TrainBuilders, config: TrainConfig) -> "Trainer":
    # note for devs: you don't have to use this function to build a custom
    # trainer, you can build it however you like. This is here for convenience.
    logging.info("Initializing training data loader")
    train_data = builder.get_train_data()
    logging.info("Initializing validation data loader")
    validation_data = builder.get_validation_data()
    if config.inference is None:
        logging.info("Skipping inline inference")
    else:
        logging.info("Initializing inline inference data loader")
    inference_data = builder.get_evaluation_inference_data()

    variable_metadata = get_derived_variable_metadata() | train_data.variable_metadata

    dataset_info = train_data.dataset_info
    logging.info("Starting model initialization")
    stepper = builder.get_stepper(
        dataset_info=dataset_info,
    )
    end_of_batch_ops = builder.get_end_of_batch_ops(
        modules=stepper.modules, base_weights=stepper.get_base_weights()
    )

    if config.inference is None:
        initial_inference_times = None
    else:
        initial_inference_times = inference_data.initial_time
    inference_n_forward_steps = config.inference_n_forward_steps
    record_step_20 = inference_n_forward_steps >= 20

    aggregator_builder = AggregatorBuilder(
        inference_config=config.inference_aggregator,
        dataset_info=dataset_info.update_variable_metadata(variable_metadata),
        output_dir=config.output_dir,
        initial_inference_time=initial_inference_times,
        record_step_20=record_step_20,
        n_timesteps=inference_n_forward_steps + stepper.n_ic_timesteps,
        loss_scaling=stepper.effective_loss_scaling,
        channel_mean_names=stepper.loss_names,
        normalize=stepper.normalizer.normalize,
        save_per_epoch_diagnostics=config.save_per_epoch_diagnostics,
        validation_config=config.validation_aggregator,
    )
    do_gc_collect = fme.get_device() != torch.device("cpu")
    trainer_config: TrainConfigProtocol = config  # documenting trainer input type
    trainer = Trainer(
        train_data=train_data,
        validation_data=validation_data,
        inference_data=inference_data,
        stepper=stepper,
        build_optimization=builder.get_optimization,
        build_ema=builder.get_ema,
        config=trainer_config,
        aggregator_builder=aggregator_builder,
        end_of_batch_callback=end_of_batch_ops,
        do_gc_collect=do_gc_collect,
    )

    def inference(
        data: InferenceDataABC[PrognosticState, BatchData],
        aggregator: InferenceEvaluatorAggregator,
        label: str,
        epoch: int,
    ):
        logging.info("Starting weather evaluation inference run")
        return inference_one_epoch(
            stepper=stepper,
            validation_context=trainer.validation_context,
            dataset=data,
            aggregator=aggregator,
            label=label,
            epoch=epoch,
        )

    end_of_epoch_ops = builder.get_end_of_epoch_callback(
        inference,
        normalize=stepper.normalizer.normalize,
        channel_mean_names=stepper.loss_names,
        output_dir=config.output_dir,
        variable_metadata=variable_metadata,
        save_diagnostics=config.save_per_epoch_diagnostics,
        n_ic_timesteps=stepper.n_ic_timesteps,
    )
    trainer.set_end_of_epoch_callback(end_of_epoch_ops)
    return trainer


class AggregatorBuilder(
    AggregatorBuilderABC[PrognosticState, TrainOutput, PairedData],
):
    def __init__(
        self,
        inference_config: InferenceEvaluatorAggregatorConfig | None,
        dataset_info: DatasetInfo,
        initial_inference_time: xr.DataArray | None,
        record_step_20: bool,
        n_timesteps: int,
        output_dir: str,
        normalize: Callable[[TensorMapping], TensorDict],
        loss_scaling: dict[str, torch.Tensor] | None = None,
        channel_mean_names: Sequence[str] | None = None,
        save_per_epoch_diagnostics: bool = False,
        validation_config: OneStepAggregatorConfig = dataclasses.field(
            default_factory=lambda: OneStepAggregatorConfig(),
        ),
    ):
        self.inference_config = inference_config
        self.dataset_info = dataset_info
        self.initial_inference_time = initial_inference_time
        self.record_step_20 = record_step_20
        self.n_timesteps = n_timesteps
        self.loss_scaling = loss_scaling
        self.channel_mean_names = channel_mean_names
        self.normalize = normalize
        self.output_dir = output_dir
        self.save_per_epoch_diagnostics = save_per_epoch_diagnostics
        self.validation_config = validation_config

    def get_train_aggregator(self) -> TrainAggregator:
        return TrainAggregator()

    def get_validation_aggregator(self) -> OneStepAggregator:
        return self.validation_config.build(
            dataset_info=self.dataset_info,
            loss_scaling=self.loss_scaling,
            save_diagnostics=self.save_per_epoch_diagnostics,
            output_dir=os.path.join(self.output_dir, "val"),
        )

    def get_inference_aggregator(
        self,
    ) -> InferenceEvaluatorAggregator:
        if isinstance(self.inference_config, InferenceEvaluatorAggregatorConfig):
            return self.inference_config.build(
                dataset_info=self.dataset_info,
                initial_time=self.initial_inference_time,
                record_step_20=self.record_step_20,
                n_timesteps=self.n_timesteps,
                channel_mean_names=self.channel_mean_names,
                normalize=self.normalize,
                save_diagnostics=self.save_per_epoch_diagnostics,
                output_dir=os.path.join(self.output_dir, "inference"),
            )
        else:
            raise ValueError(
                "Trying to build an inference aggregator, but inference config not set."
            )


def run_train_from_config(config: TrainConfig):
    run_train(TrainBuilders(config), config)


def run_train(builders: TrainBuilders, config: TrainConfig):
    dist = Distributed.get_instance()
    if fme.using_gpu():
        torch.backends.cudnn.benchmark = True
    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir, exist_ok=True)
    config.logging.configure_logging(config.experiment_dir, log_filename="out.log")
    env_vars = logging_utils.retrieve_env_vars()
    logging_utils.log_versions()
    beaker_url = logging_utils.log_beaker_url()
    config_as_dict = to_flat_dict(dataclasses.asdict(config))
    config.logging.configure_wandb(
        config=config_as_dict,
        env_vars=env_vars,
        notes=beaker_url,
    )
    trainer = build_trainer(builders, config)
    trainer.train()
    logging.info(f"DONE ---- rank {dist.rank}")


def main(yaml_config: str, override_dotlist: Sequence[str] | None = None):
    config_data = prepare_config(yaml_config, override=override_dotlist)
    config = dacite.from_dict(
        data_class=TrainConfig, data=config_data, config=dacite.Config(strict=True)
    )
    prepare_directory(config.experiment_dir, config_data)
    run_train_from_config(config)



import warnings
from collections.abc import Sequence

import numpy as np
import xarray as xr

from fme.core.dataset.concat import XarrayConcat
from fme.core.dataset.config import (
    ConcatDatasetConfig,
    MergeDatasetConfig,
    MergeNoConcatDatasetConfig,
    RepeatedInterval,
    TimeSlice,
    XarrayDataConfig,
)
from fme.core.dataset.merged import MergedXarrayDataset
from fme.core.dataset.properties import DatasetProperties
from fme.core.dataset.subset import XarraySubset
from fme.core.dataset.xarray import XarrayDataset, get_raw_paths
from fme.core.typing_ import Slice


def _as_index_selection(
    subset: Slice | TimeSlice | RepeatedInterval, dataset: XarrayDataset
) -> slice | np.ndarray:
    """Converts a subset defined either as a Slice or TimeSlice into an index slice
    based on time coordinate in provided dataset.
    """
    if isinstance(subset, Slice):
        index_selection = subset.slice
    elif isinstance(subset, TimeSlice):
        index_selection = subset.slice(dataset.sample_start_times)
    elif isinstance(subset, RepeatedInterval):
        try:
            index_selection = subset.get_boolean_mask(len(dataset), dataset.timestep)
        except ValueError as e:
            raise ValueError(f"Error when applying RepeatedInterval to dataset: {e}")
    else:
        raise TypeError(f"subset must be Slice or TimeSlice, got {type(subset)}")
    return index_selection


def get_xarray_dataset(
    config: XarrayDataConfig, names: list[str], n_timesteps: int
) -> tuple["XarraySubset", DatasetProperties]:
    dataset = XarrayDataset(config, names, n_timesteps)
    properties = dataset.properties
    index_slice = _as_index_selection(config.subset, dataset)
    dataset = XarraySubset(dataset, index_slice)
    return dataset, properties


def get_datasets(
    dataset_configs: Sequence[XarrayDataConfig],
    names: list[str],
    n_timesteps: int,
    strict: bool = True,
) -> tuple[list[XarraySubset], DatasetProperties]:
    datasets = []
    properties: DatasetProperties | None = None
    for config in dataset_configs:
        dataset, new_properties = get_xarray_dataset(config, names, n_timesteps)
        datasets.append(dataset)
        if properties is None:
            properties = new_properties
        elif not strict:
            try:
                properties.update(new_properties)
            except ValueError as e:
                warnings.warn(
                    f"Metadata for each ensemble member are not the same: {e}"
                )
        else:
            properties.update(new_properties)
    if properties is None:
        raise ValueError("At least one dataset must be provided.")

    return datasets, properties


def get_dataset(
    dataset_configs: Sequence[XarrayDataConfig],
    names: list[str],
    n_timesteps: int,
    strict: bool = True,
) -> tuple[XarrayConcat, DatasetProperties]:
    datasets, properties = get_datasets(
        dataset_configs, names, n_timesteps, strict=strict
    )
    ensemble = XarrayConcat(datasets)
    return ensemble, properties


def _infer_available_variables(config: XarrayDataConfig):
    """
    Infer the available variables from a XarrayDataset.
    """
    paths = get_raw_paths(config.data_path, config.file_pattern)
    dataset = xr.open_dataset(
        paths[0],
        decode_times=False,
        decode_timedelta=False,
        engine=config.engine,
        chunks=None,
    )
    return dataset.data_vars


def get_per_dataset_names(
    merged_config: MergeDatasetConfig | MergeNoConcatDatasetConfig,
    names: list[str],
) -> list[list[str]]:
    merged_required_names = names.copy()
    per_dataset_names = []
    for config in merged_config.merge:
        if isinstance(config, XarrayDataConfig):
            current_source_variables = _infer_available_variables(config)
        elif isinstance(config, ConcatDatasetConfig):
            current_source_variables = _infer_available_variables(config.concat[0])
        current_source_names = [
            name for name in merged_required_names if name in current_source_variables
        ]
        per_dataset_names.append(current_source_names)
        for name in current_source_names:
            merged_required_names.remove(name)
    return per_dataset_names


def get_merged_datasets(
    merged_config: MergeDatasetConfig,
    names: list[str],
    n_timesteps: int,
) -> tuple[MergedXarrayDataset, DatasetProperties]:
    merged_xarray_datasets = []
    merged_properties: DatasetProperties | None = None
    per_dataset_names = get_per_dataset_names(merged_config, names)
    config_counter = 0
    for config in merged_config.merge:
        if isinstance(config, XarrayDataConfig):
            current_source_xarray_dataset, current_source_properties = (
                get_xarray_dataset(
                    config,
                    per_dataset_names[config_counter],
                    n_timesteps,
                )
            )
            merged_xarray_datasets.append(current_source_xarray_dataset)
        elif isinstance(config, ConcatDatasetConfig):
            current_source_datasets, current_source_properties = get_datasets(
                config.concat,
                per_dataset_names[config_counter],
                n_timesteps,
                strict=config.strict,
            )
            current_source_ensemble = XarrayConcat(current_source_datasets)
            merged_xarray_datasets.append(current_source_ensemble)

        if merged_properties is None:
            merged_properties = current_source_properties
        else:
            merged_properties.update_merged_dataset(current_source_properties)
        config_counter += 1

    if merged_properties is None:
        raise ValueError("At least one dataset must be provided.")
    merged_datasets = MergedXarrayDataset(datasets=merged_xarray_datasets)
    return merged_datasets, merged_properties
